""" Test for get_index function in eigenstates.py
"""

from excitingtools.eigenstates.eigenstates import get_k_point_index
import pytest
import numpy as np
import re


def test_get_k_point_index():
    k_points = np.array([[1.000000, 0.000000, 0.000000],
                         [0.988281, 0.011719, 0.000000],
                         [0.988281, 0.011719, 0.000000],
                         [0.976562, 0.023438, 0.000000],
                         [0.953125, 0.046875, 0.000000]])

    assert get_k_point_index([1.000000, 0.000000, 0.000000], k_points) == 0
    assert get_k_point_index([0.953125, 0.046875, 0.000000], k_points) == 4
    assert np.isnan(get_k_point_index([0.000000, 0.500000, 0.500000], k_points)), "No k-point matched"
    with pytest.raises(ValueError, match=re.escape("Found degenerate k-points at [[1], [2]]")):
        get_k_point_index([0.988281, 0.011719, 0.000000], k_points), "ValueError is returned for degenerate k-points"



""" Automatically generated file with the valid attributes from the schema. 
Do not manually change. Instead, run "utils/schema_parsing.py" to regenerate. """ 

# input information 
input_valid_attributes = ['sharedfs', 'xsltpath'] 
input_valid_subtrees = ['title', 'structure', 'groundstate', 'relax', 'properties', 'phonons', 'xs', 'gw', 'MD', 'eph', 
                        'keywords'] 
input_mandatory_attributes = ['groundstate', 'structure', 'title'] 


# common information 
origin_valid_attributes = ['coord'] 

point_valid_attributes = ['breakafter', 'coord', 'label'] 
point_mandatory_attributes = ['coord'] 

plot1d_valid_subtrees = ['path'] 
plot1d_mandatory_attributes = ['path'] 

path_valid_attributes = ['outfileprefix', 'steps'] 
path_valid_subtrees = ['point'] 
path_mandatory_attributes = ['point', 'steps'] 
path_multiple_children = ['point'] 

plot2d_valid_subtrees = ['parallelogram'] 
plot2d_mandatory_attributes = ['parallelogram'] 

parallelogram_valid_attributes = ['grid', 'outfileprefix'] 
parallelogram_valid_subtrees = ['origin', 'point'] 
parallelogram_mandatory_attributes = ['grid', 'origin', 'point'] 
parallelogram_multiple_children = ['point'] 

plot3d_valid_attributes = ['usesym'] 
plot3d_valid_subtrees = ['box'] 
plot3d_mandatory_attributes = ['box'] 

box_valid_attributes = ['grid', 'outfileprefix'] 
box_valid_subtrees = ['origin', 'point'] 
box_mandatory_attributes = ['grid', 'origin', 'point'] 
box_multiple_children = ['point'] 

kstlist_valid_subtrees = ['pointstatepair'] 
kstlist_mandatory_attributes = ['pointstatepair'] 
kstlist_multiple_children = ['pointstatepair'] 

energywindow_valid_attributes = ['intv', 'points'] 

qpointset_valid_subtrees = ['qpoint'] 
qpointset_mandatory_attributes = ['qpoint'] 
qpointset_multiple_children = ['qpoint'] 

parts_valid_subtrees = ['dopart'] 
parts_multiple_children = ['dopart'] 

dopart_valid_attributes = ['id'] 
dopart_mandatory_attributes = ['id'] 


# structure information 
structure_valid_attributes = ['autormt', 'autormtscaling', 'cartesian', 'epslat', 'primcell', 'speciespath', 'tshift'] 
structure_valid_subtrees = ['crystal', 'species', 'symmetries'] 
structure_mandatory_attributes = ['speciespath'] 
structure_multiple_children = ['species'] 

crystal_valid_attributes = ['scale', 'stretch'] 
crystal_valid_subtrees = ['basevect'] 
crystal_multiple_children = ['basevect'] 

species_valid_attributes = ['atomicNumber', 'chemicalSymbol', 'fixrmt', 'rmt', 'speciesfile'] 
species_valid_subtrees = ['atom', 'LDAplusU', 'dfthalfparam'] 
species_mandatory_attributes = ['speciesfile'] 
species_multiple_children = ['atom'] 

atom_valid_attributes = ['bfcmt', 'coord', 'lockxyz', 'mommtfix', 'velocity'] 
atom_mandatory_attributes = ['coord'] 

LDAplusU_valid_attributes = ['J', 'U', 'l'] 

dfthalfparam_valid_attributes = ['ampl', 'cut', 'exponent'] 
dfthalfparam_valid_subtrees = ['shell'] 
dfthalfparam_mandatory_attributes = ['shell'] 
dfthalfparam_multiple_children = ['shell'] 

shell_valid_attributes = ['ionization', 'number'] 


# groundstate information 
groundstate_valid_attributes = ['APWprecision', 'CoreRelativity', 'ExplicitKineticEnergy', 'PrelimLinSteps', 
                                'ValenceRelativity', 'autokpt', 'beta0', 'betadec', 'betainc', 'cfdamp', 'chgexs', 
                                'deband', 'dipolecorrection', 'dipoleposition', 'dlinengyfermi', 'do', 'energyref', 
                                'epsband', 'epschg', 'epsengy', 'epsforcescf', 'epsocc', 'epspot', 'fermilinengy', 
                                'findlinentype', 'fracinr', 'frozencore', 'gmaxvr', 'isgkmax', 'ldapu', 'lmaxapw', 
                                'lmaxinr', 'lmaxmat', 'lmaxvr', 'lradstep', 'maxscl', 'mixer', 'mixerswitch', 
                                'modifiedsv', 'msecStoredSteps', 'nempty', 'ngridk', 'niterconvcheck', 'nktot', 
                                'nosource', 'nosym', 'nprad', 'npsden', 'nwrite', 'outputlevel', 'ptnucl', 
                                'radialgridtype', 'radkpt', 'reducek', 'rgkmax', 'scfconv', 'stype', 'swidth', 
                                'symmorph', 'tevecsv', 'tfibs', 'tforce', 'tpartcharges', 'useAPWprecision', 
                                'useDensityMatrix', 'vdWcorrection', 'vkloff', 'xctype'] 
groundstate_valid_subtrees = ['DFTD2parameters', 'TSvdWparameters', 'spin', 'HartreeFock', 'dfthalf', 'Hybrid', 
                              'sirius', 'solver', 'OEP', 'RDMFT', 'output', 'libxc', 'xsLO', 'lorecommendation'] 

DFTD2parameters_valid_attributes = ['cutoff', 'd', 's6', 'sr6'] 

TSvdWparameters_valid_attributes = ['cutoff', 'd', 'nr', 'nsph', 's6', 'sr6'] 

spin_valid_attributes = ['bfieldc', 'fixspin', 'momfix', 'nosv', 'realspace', 'reducebf', 'spinorb', 'spinsprl', 'svlo', 
                         'taufsm', 'vqlss'] 

dfthalf_valid_attributes = ['printVSfile'] 

Hybrid_valid_attributes = ['HSEsingularity', 'eccoeff', 'epsmb', 'exchangetype', 'excoeff', 'gmb', 'lmaxmb', 'maxscl', 
                           'mblksiz', 'omega', 'updateRadial'] 

sirius_valid_attributes = ['cfun', 'density', 'densityinit', 'eigenstates', 'sfacg', 'vha', 'xc'] 

solver_valid_attributes = ['constructHS', 'evaltol', 'minenergy', 'packedmatrixstorage', 'type'] 

OEP_valid_attributes = ['convoep', 'maxitoep', 'tauoep'] 

RDMFT_valid_attributes = ['maxitc', 'maxitn', 'rdmalpha', 'rdmmaxscl', 'rdmtemp', 'rdmxctype', 'taurdmc', 'taurdmn'] 

output_valid_attributes = ['state'] 

libxc_valid_attributes = ['correlation', 'exchange', 'xc'] 

xsLO_valid_attributes = ['emax', 'lmax', 'maxnodes'] 

lorecommendation_valid_attributes = ['lmaxlo', 'nodesmaxlo'] 


# relax information 
relax_valid_attributes = ['addtohistory', 'endbfgs', 'epsforce', 'history', 'historyformat', 'maxbfgs', 'maxsteps', 
                          'method', 'outputlevel', 'printtorque', 'taubfgs', 'taunewton'] 


# phonons information 
phonons_valid_attributes = ['canonical', 'delete_eigensystem_response', 'deltaph', 'do', 'drynumprocs', 'gamma', 
                            'maxprocsperpart', 'method', 'minprocsperpart', 'ngridq', 'polar', 'reduceq', 'sumrule', 
                            'write_schedule'] 
phonons_valid_subtrees = ['qpointset', 'phonondos', 'phonondispplot', 'reformatdynmat', 'interpolate', 'parts'] 

phonondos_valid_attributes = ['inttype', 'ngrdos', 'ngridqint', 'nsmdos', 'ntemp', 'nwdos'] 

phonondispplot_valid_subtrees = ['plot1d'] 
phonondispplot_mandatory_attributes = ['plot1d'] 

interpolate_valid_attributes = ['ngridq', 'vqloff', 'writeeigenvectors'] 
interpolate_mandatory_attributes = ['ngridq'] 


# properties information 
properties_valid_subtrees = ['spintext', 'coreoverlap', 'bandstructure', 'stm', 'wfplot', 'dos', 'LSJ', 'masstensor', 
                             'chargedensityplot', 'TSvdW', 'DFTD2', 'exccplot', 'elfplot', 'mvecfield', 'xcmvecfield', 
                             'electricfield', 'gradmvecfield', 'fermisurfaceplot', 'EFG', 'mossbauer', 'expiqr', 
                             'elnes', 'eliashberg', 'momentummatrix', 'dielmat', 'boltzequ', 'raman', 'moke', 'shg', 
                             'wannier', 'wannierplot', 'wanniergap', 'ldos', 'polarization'] 

spintext_valid_attributes = ['bands'] 
spintext_valid_subtrees = ['plot2d'] 
spintext_mandatory_attributes = ['plot2d'] 

coreoverlap_valid_attributes = ['coreatom', 'corespecies'] 

bandstructure_valid_attributes = ['character', 'deriv', 'scissor', 'wannier'] 
bandstructure_valid_subtrees = ['plot1d'] 
bandstructure_mandatory_attributes = ['plot1d'] 

stm_valid_attributes = ['bias', 'stmmode', 'stmtype'] 
stm_valid_subtrees = ['plot2d', 'region'] 

region_valid_attributes = ['grid2d', 'grid3d', 'height', 'zrange'] 

wfplot_valid_attributes = ['version'] 
wfplot_valid_subtrees = ['kstlist', 'plot1d', 'plot2d', 'plot3d'] 
wfplot_mandatory_attributes = ['kstlist'] 

dos_valid_attributes = ['inttype', 'jdos', 'linkpt', 'lmirep', 'lonly', 'ngrdos', 'ngridkint', 'nsmdos', 'nwdos', 
                        'scissor', 'sqados', 'wannier', 'winddos'] 

LSJ_valid_subtrees = ['kstlist'] 

masstensor_valid_attributes = ['deltaem', 'ndspem', 'vklem'] 

chargedensityplot_valid_attributes = ['nocore'] 
chargedensityplot_valid_subtrees = ['plot1d', 'plot2d', 'plot3d'] 

exccplot_valid_subtrees = ['plot1d', 'plot2d', 'plot3d'] 

elfplot_valid_subtrees = ['plot1d', 'plot2d', 'plot3d'] 

mvecfield_valid_subtrees = ['plot2d', 'plot3d'] 

xcmvecfield_valid_subtrees = ['plot2d', 'plot3d'] 

electricfield_valid_subtrees = ['plot2d', 'plot3d'] 

gradmvecfield_valid_subtrees = ['plot1d', 'plot2d', 'plot3d'] 

fermisurfaceplot_valid_attributes = ['nstfsp'] 
fermisurfaceplot_valid_subtrees = ['plot2d', 'plot3d'] 

expiqr_valid_subtrees = ['kstlist'] 

elnes_valid_attributes = ['ngrid', 'vecql', 'wgrid', 'wmax', 'wmin'] 

eliashberg_valid_attributes = ['mustar'] 

momentummatrix_valid_attributes = ['fastpmat'] 

dielmat_valid_attributes = ['drude', 'intraband', 'scissor', 'swidth', 'tevout', 'wgrid', 'wmax'] 
dielmat_valid_subtrees = ['epscomp'] 
dielmat_multiple_children = ['epscomp'] 

boltzequ_valid_attributes = ['chemicalPotentialRange', 'chemicalPotentialSpacing', 'dopingConcentration', 
                             'energyReference', 'evOutputEnergies', 'siOutputUnits', 'temperatureRange', 
                             'temperatureSpacing', 'transportDfBroadening', 'transportDfRange', 'transportDfSpacing', 
                             'useDopingConcentration', 'useTransportDf'] 
boltzequ_valid_subtrees = ['etCoeffComponents'] 
boltzequ_multiple_children = ['etCoeffComponents'] 

raman_valid_attributes = ['broad', 'degree', 'displ', 'doequilibrium', 'elaser', 'elaserunit', 'getphonon', 'mode', 
                          'molecule', 'ninter', 'nstate', 'nstep', 'temp', 'useforces', 'usesym', 'writefunc', 'xmax', 
                          'xmin'] 
raman_valid_subtrees = ['eigvec', 'energywindow'] 
raman_mandatory_attributes = ['energywindow'] 
raman_multiple_children = ['eigvec'] 

eigvec_valid_attributes = ['comp'] 
eigvec_mandatory_attributes = ['comp'] 

moke_valid_attributes = ['drude', 'intraband', 'scissor', 'swidth', 'tevout', 'wgrid', 'wmax'] 

shg_valid_attributes = ['etol', 'scissor', 'swidth', 'tevout', 'wgrid', 'wmax'] 
shg_valid_subtrees = ['chicomp'] 
shg_mandatory_attributes = ['chicomp'] 
shg_multiple_children = ['chicomp'] 

wannier_valid_attributes = ['cutshell', 'do', 'fermizero', 'input', 'mindist', 'minshell', 'nbzshell', 'printproj'] 
wannier_valid_subtrees = ['projection', 'group'] 
wannier_multiple_children = ['group'] 

projection_valid_attributes = ['dordmax', 'epsld', 'nprojtot', 'nunocc'] 

group_valid_attributes = ['epsdis', 'epsmax', 'epsopf', 'epsproj', 'fst', 'innerwindow', 'lst', 'maxitdis', 'maxitmax', 
                          'maxitopf', 'memlendis', 'memlenmax', 'memlenopf', 'method', 'minitdis', 'minitmax', 
                          'minitopf', 'minstepdis', 'minstepmax', 'minstepopf', 'neighcells', 'nproj', 'nwf', 'nwrite', 
                          'optim', 'outerwindow', 'writeconv'] 
group_valid_subtrees = ['projector'] 
group_multiple_children = ['projector'] 

projector_valid_attributes = ['nr'] 
projector_mandatory_attributes = ['nr'] 

wannierplot_valid_attributes = ['cell', 'fst', 'lst'] 
wannierplot_valid_subtrees = ['plot1d', 'plot2d', 'plot3d'] 

wanniergap_valid_attributes = ['auto', 'ngridkint'] 
wanniergap_valid_subtrees = ['pointband'] 
wanniergap_multiple_children = ['pointband'] 

pointband_valid_attributes = ['band', 'extremal', 'vkl'] 
pointband_mandatory_attributes = ['band', 'vkl'] 

ldos_valid_attributes = ['delta', 'grid', 'newint', 'ngrdos', 'nsmdos', 'nwdos', 'scissor', 'tol', 'winddos'] 


# xs information 
xs_valid_attributes = ['bfieldc', 'broad', 'dbglev', 'dfoffdiag', 'dogroundstate', 'emattype', 'emaxdf', 'epsdfde', 
                       'fastpmat', 'gqmax', 'gqmaxtype', 'h5fname', 'h5gname', 'lmaxapwwf', 'lmaxemat', 'maxscl', 
                       'nempty', 'ngridk', 'ngridq', 'nosym', 'pwmat', 'reducek', 'reduceq', 'rgkmax', 'scissor', 
                       'skipgnd', 'swidth', 'tappinfo', 'tevout', 'vkloff', 'writexsgrids', 'xstype'] 
xs_valid_subtrees = ['storeexcitons', 'pwelements', 'writeexcitons', 'writekpathweights', 'excitonPlot', 
                     'realTimeTDDFT', 'tddft', 'screening', 'phonon_screening', 'expand_eps', 'BSE', 'fastBSE', 
                     'transitions', 'qpointset', 'tetra', 'energywindow', 'plan'] 
xs_mandatory_attributes = ['xstype'] 

storeexcitons_valid_attributes = ['MaxEnergyExcitons', 'MaxNumberExcitons', 'MinEnergyExcitons', 'MinNumberExcitons', 
                                  'selectenergy', 'useev'] 

pwelements_valid_attributes = ['band_combinations'] 
pwelements_mandatory_attributes = ['band_combinations'] 

writeexcitons_valid_attributes = ['MaxEnergyExcitons', 'MaxNumberExcitons', 'MinEnergyExcitons', 'MinNumberExcitons', 
                                  'abscutares', 'abscutres', 'selectenergy', 'useev'] 

writekpathweights_valid_attributes = ['MaxEnergyExcitons', 'MaxNumberExcitons', 'MinEnergyExcitons', 
                                      'MinNumberExcitons', 'intorder', 'printgridweights', 'selectenergy', 'useev'] 

excitonPlot_valid_attributes = ['epstol'] 
excitonPlot_valid_subtrees = ['exciton', 'hole', 'electron'] 
excitonPlot_mandatory_attributes = ['electron', 'hole'] 
excitonPlot_multiple_children = ['exciton'] 

exciton_valid_attributes = ['fix', 'lambda'] 

hole_valid_subtrees = ['plot1d', 'plot2d', 'plot3d'] 

electron_valid_subtrees = ['plot1d', 'plot2d', 'plot3d'] 

realTimeTDDFT_valid_attributes = ['TaylorOrder', 'calculateNExcitedElectrons', 'calculateTotalEnergy', 'endTime', 
                                  'forcePmatHermitian', 'normalizeWF', 'printAfterIterations', 'printTimingDetailed', 
                                  'printTimingGeneral', 'propagator', 'readPmatBasis', 'subtractJ0', 'timeStep', 
                                  'vectorPotentialSolver'] 
realTimeTDDFT_valid_subtrees = ['predictorCorrector', 'screenshots', 'laser'] 

predictorCorrector_valid_attributes = [' maxIterations', 'tol'] 

screenshots_valid_attributes = ['niter', 'printAbsProjCoeffs'] 

laser_valid_attributes = ['fieldType'] 
laser_valid_subtrees = ['kick', 'trapCos', 'sinSq'] 
laser_multiple_children = ['kick', 'sinSq', 'trapCos'] 

kick_valid_attributes = ['amplitude', 'direction', 't0', 'width'] 

trapCos_valid_attributes = ['amplitude', 'direction', 'omega', 'phase', 'riseTime', 't0', 'width'] 

sinSq_valid_attributes = ['amplitude', 'direction', 'omega', 'phase', 'pulseLength', 't0'] 

tddft_valid_attributes = ['acont', 'ahc', 'alphalrc', 'alphalrcdyn', 'aresdf', 'aresfxc', 'betalrcdyn', 'do', 'drude', 
                          'fxcbsesplit', 'fxctype', 'intraband', 'kerndiag', 'lindhard', 'lmaxalda', 'mdfqtype', 
                          'nwacont', 'torddf', 'tordfxc'] 

screening_valid_attributes = ['do', 'intraband', 'nempty', 'ngridk', 'ngridq_interpolation', 'nosym', 'nqpt_unique', 
                              'qpointsgamma', 'quasiparticle_correction', 'reducek', 'rgkmax', 'screentype', 'tr', 
                              'vkloff'] 

phonon_screening_valid_attributes = ['alat_qe', 'excitation_energy', 'file_type', 'phonon_file', 'zstar_file'] 
phonon_screening_mandatory_attributes = ['alat_qe', 'excitation_energy', 'phonon_file', 'zstar_file'] 

expand_eps_valid_attributes = ['supercell_1', 'supercell_2'] 

BSE_valid_attributes = ['aresbse', 'blocks', 'bsedirsing', 'bsetype', 'checkposdef', 'chibar0', 'chibar0comp', 
                        'chibarq', 'coupling', 'cuttype', 'dichroism', 'distribute', 'econv', 'eecs', 'efind', 'fbzq', 
                        'iqmtrange', 'lmaxdielt', 'measure', 'nexc', 'ngridksub', 'nleblaik', 'nosym', 'nosymspec', 
                        'nstlbse', 'nstlxas', 'outputlevel', 'reducek', 'rgkmax', 'sciavbd', 'sciavqbd', 'sciavqhd', 
                        'sciavqwg', 'sciavtype', 'scrherm', 'solver', 'vkloff', 'writehamhdf5', 'writepotential', 'xas', 
                        'xasatom', 'xasedge', 'xasspecies', 'xes'] 

fastBSE_valid_attributes = ['cvtsteplim', 'cvttol', 'lanczosmaxits', 'nisdf', 'rsampling', 'seed'] 

transitions_valid_subtrees = ['individual', 'ranges', 'lists'] 

individual_valid_subtrees = ['trans'] 
individual_multiple_children = ['trans'] 

trans_valid_attributes = ['action', 'final', 'initial', 'kpointnumber'] 

ranges_valid_subtrees = ['range'] 
ranges_multiple_children = ['range'] 

range_valid_attributes = ['action', 'kpointnumber', 'start', 'statestype', 'stop'] 
range_mandatory_attributes = ['statestype'] 

lists_valid_subtrees = ['istate'] 
lists_multiple_children = ['istate'] 

istate_valid_attributes = ['action', 'kpointnumber', 'state', 'statestype'] 
istate_mandatory_attributes = ['statestype'] 

tetra_valid_attributes = ['cw1k', 'kordexc', 'qweights', 'tetradf', 'tetraocc'] 

plan_valid_subtrees = ['doonly'] 
plan_multiple_children = ['doonly'] 

doonly_valid_attributes = ['task'] 
doonly_mandatory_attributes = ['task'] 


# gw information 
gw_valid_attributes = ['at1', 'at2', 'coreflag', 'debug', 'eph', 'ibgw', 'ibmax', 'ibmax2', 'ibmin', 'ibmin2', 'igmax', 
                       'igmin', 'iik', 'jjk', 'mblksiz', 'nbgw', 'nempty', 'ngridq', 'printSelfC', 
                       'printSpectralFunction', 'qdepw', 'reduceq', 'rmax', 'rpath', 'rpmat', 'skipgnd', 'taskname', 
                       'vqloff', 'wlo', 'wto'] 
gw_valid_subtrees = ['plot1d', 'freqgrid', 'selfenergy', 'mixbasis', 'barecoul', 'scrcoul'] 

freqgrid_valid_attributes = ['eta', 'fconv', 'fgrid', 'freqmax', 'freqmin', 'nomeg'] 

selfenergy_valid_attributes = ['actype', 'eqpsolver', 'eshift', 'method', 'nempty', 'singularity', 'swidth', 'tol'] 
selfenergy_valid_subtrees = ['wgrid'] 

wgrid_valid_attributes = ['size', 'type', 'wmax', 'wmin'] 

mixbasis_valid_attributes = ['epsmb', 'gmb', 'lmaxmb'] 

barecoul_valid_attributes = ['barcevtol', 'basis', 'cutofftype', 'pwm', 'stctol'] 

scrcoul_valid_attributes = ['averaging', 'omegap', 'q0eps', 'scrtype'] 


# MD information 
MD_valid_attributes = ['basisDerivative', 'coreCorrections', 'integrationAlgorithm', 'printAllForces', 'timeStep', 
                       'type', 'updateOverlap', 'updatePmat', 'valenceCorrections'] 


# eph information 
eph_valid_attributes = ['debugeph', 'ibeph', 'ibsumeph', 'nbeph', 'nbsumeph', 'nemptyeph', 'ngridqeph', 'tasknameeph', 
                        'vqloffeph'] 
eph_valid_subtrees = ['freqgrideph', 'selfenergyeph'] 

freqgrideph_valid_attributes = ['freqmaxeph', 'nomegeph'] 

selfenergyeph_valid_subtrees = ['SpectralFunctionPloteph'] 

SpectralFunctionPloteph_valid_attributes = ['axis', 'eta', 'nwgrid', 'wmax', 'wmin'] 


# valid entries for the xs subtree 'plan'
valid_plan_entries = ['bse', 'bsegenspec', 'bsesurvey', 'df', 'df2', 'dielectric', 'emattest', 'exccoulint', 
                      'excitonWavefunction', 'expand_add_eps', 'fastBSE_isdf_cvt', 'fastBSE_isdf_vexc_test', 
                      'fastBSE_main', 'fastBSE_setup_transitions', 'fxc_alda_check', 'idf', 'kernxc_bse', 'kernxc_bse3', 
                      'phonon_screening', 'planewave_elements', 'pmatxs2orig', 'portstate(-1)', 'portstate(-2)', 
                      'portstate(1)', 'portstate(2)', 'scrcoulint', 'screen', 'scrgeneigvec', 'scrtetcalccw', 
                      'scrwritepmat', 'testmain', 'testxs', 'tetcalccw', 'write_dielectric_matrix', 
                      'write_pmat_hdf5_xs', 'write_screen', 'write_screened_coulomb', 'write_wfplot', 
                      'writebandgapgrid', 'writebevec', 'writeemat', 'writeematasc', 'writekpathweights', 
                      'writeoverlapxs', 'writepmat', 'writepmatasc', 'writepmatxs', 'writepwmat', 'x0toasc', 'x0tobin', 
                      'xsestimate', 'xsgeneigvec'] 



from excitingtools.exciting_dict_parsers.state_parser import parse_state_out
from pathlib import Path
import numpy as np
import sys


def write_int(file, i: int, byteorder):
    """ Writes an integer as bytes to the file.

    :param file: file object to write the int
    :param i: integer, which bytes should be written
    :param byteorder: endianness of the byteorder ("little" or "big")
    """
    file.write(int.to_bytes(4, 4, byteorder))
    file.write(i.to_bytes(4, byteorder))
    file.write(int.to_bytes(4, 4, byteorder))


def write_array(file, a: np.ndarray, byteorder):
    """ Writes a numpy array to the file as bytes.

    :param file: file object to write the int
    :param a: numpy array, which bytes should be written
    :param byteorder: endianness of the byteorder ("little" or "big")
    """
    num_bytes = a.size * 8
    file.write(num_bytes.to_bytes(4, byteorder))
    file.write(a.tobytes(order='F'))
    file.write(num_bytes.to_bytes(4, byteorder))


def write_state(path, state: dict):
    """ Creates a file at the specified path and writes the dictionary in the same structure as STATE.OUT

    :param path: path to the file to which should be written to
    :param state: state dictionary, which should be converted to a binary STATE file
    """
    byteorder = sys.byteorder
    with open(path, "wb+") as file:
        # write version information
        file.write(int.to_bytes(52, 4, byteorder))
        for i in state["version"]:
            i: int
            file.write(i.to_bytes(4, byteorder))
        file.write("YomNFS0t4s8uhGxs4FCNGsAOp5DapSm7HAQ58aVd".encode("utf-8"))
        file.write(int.to_bytes(52, 4, byteorder))
        # write other general information
        for key in ("spinpol", "nspecies", "lmmaxvr", "nrmtmax"):
            write_int(file, state[key], byteorder)
        # next write the species specific information
        for i in range(state["nspecies"]):
            write_int(file, state["number of atoms"][i], byteorder)
            write_int(file, len(state["muffintin radial meshes"][i]), byteorder)
            write_array(file, state["muffintin radial meshes"][i], byteorder)
        # next write g vector grid
        file.write(int.to_bytes(12, 4, byteorder))
        for i in range(3):
            file.write(state["g vector grid"][i].to_bytes(4, byteorder))
        file.write(int.to_bytes(12, 4, byteorder))
        # next are again a few integers
        for key in ("ngvec", "ndmag", "nspinor", "ldapu", "lmmaxlu"):
            write_int(file, state[key], byteorder)
        # next write the array pairs
        arrays = (
            ("muffintin density", "interstitial density"),
            ("muffintin coulomb potential", "interstitial coulomb potential"),
            ("muffintin exchange-correlation potential", "interstitial exchange-correlation potential"),
        )
        for muffintin_array, interstitial_array in arrays:
            muffintin_array: np.ndarray = state[muffintin_array]
            interstitial_array: np.ndarray = state[interstitial_array]
            num_bytes = (muffintin_array.size + interstitial_array.size) * 8
            file.write(num_bytes.to_bytes(4, byteorder))
            file.write(muffintin_array.tobytes(order='F'))
            file.write(interstitial_array.tobytes(order='F'))
            file.write(num_bytes.to_bytes(4, byteorder))
        # Lastly write the potential arrays
        num_bytes = 8 * (state["muffintin effective potential"].size + state["interstitial effective potential"].size +
                         2 * state["reciprocal interstitial effective potential"].size)
        file.write(num_bytes.to_bytes(4, byteorder))
        file.write(state["muffintin effective potential"].tobytes(order='F'))
        file.write(state["interstitial effective potential"].tobytes(order='F'))
        file.write(state["reciprocal interstitial effective potential"].tobytes(order='F'))
        file.write(num_bytes.to_bytes(4, byteorder))


def test_parse_state_out(tmp_path: Path):
    directory = tmp_path

    # setting seed to avoid randomness in tests
    np.random.seed(0)

    state_ref = {
        "version": (1, 2, 3),
        "versionhash": "YomNFS0t4s8uhGxs4FCNGsAOp5DapSm7HAQ58aVd",
        "spinpol": False,
        "nspecies": 2,
        "lmmaxvr": 3,
        "nrmtmax": 3,
        "number of atoms": [1, 1],
        "muffintin radial meshes": [np.array([1.0, 2.0]), np.array([0.5, 0.75, 1.5])],
        "g vector grid": (2, 2, 2),
        "ngvec": 5,
        "ndmag": 0,
        "nspinor": 1,
        "ldapu": 0,
        "lmmaxlu": 16,
        "muffintin density": np.random.random((3, 3, 2)),
        "interstitial density": np.random.random(8),
        "muffintin coulomb potential": np.random.random((3, 3, 2)),
        "interstitial coulomb potential": np.random.random(8),
        "muffintin exchange-correlation potential": np.random.random((3, 3, 2)),
        "interstitial exchange-correlation potential": np.random.random(8),
        "muffintin effective potential": np.random.random((3, 3, 2)),
        "interstitial effective potential": np.random.random(8),
        "reciprocal interstitial effective potential": np.random.random(5) + 1j * np.random.random(5),
    }
    write_state(directory / "STATE.OUT", state_ref)
    state_out = parse_state_out(directory / "STATE.OUT")

    assert state_out["version"] == state_ref["version"]
    assert state_out["versionhash"] == state_ref["versionhash"]
    assert state_out["spinpol"] == state_ref["spinpol"]
    assert state_out["nspecies"] == state_ref["nspecies"]
    assert state_out["lmmaxvr"] == state_ref["lmmaxvr"]
    assert state_out["nrmtmax"] == state_ref["nrmtmax"]
    assert state_out["number of atoms"] == state_ref["number of atoms"]
    assert np.allclose(state_out["muffintin radial meshes"][0], state_ref["muffintin radial meshes"][0])
    assert np.allclose(state_out["muffintin radial meshes"][1], state_ref["muffintin radial meshes"][1])
    assert state_out["g vector grid"] == state_ref["g vector grid"]
    assert state_out["ngvec"] == state_ref["ngvec"]
    assert state_out["ndmag"] == state_ref["ndmag"]
    assert state_out["nspinor"] == state_ref["nspinor"]
    assert state_out["ldapu"] == state_ref["ldapu"]
    assert state_out["lmmaxlu"] == state_ref["lmmaxlu"]
    assert np.allclose(state_out["muffintin density"], state_ref["muffintin density"])
    assert np.allclose(state_out["interstitial density"], state_ref["interstitial density"])
    assert np.allclose(state_out["muffintin coulomb potential"], state_ref["muffintin coulomb potential"])
    assert np.allclose(state_out["interstitial coulomb potential"], state_ref["interstitial coulomb potential"])
    assert np.allclose(state_out["muffintin exchange-correlation potential"],
                       state_ref["muffintin exchange-correlation potential"])
    assert np.allclose(state_out["interstitial exchange-correlation potential"],
                       state_ref["interstitial exchange-correlation potential"])
    assert np.allclose(state_out["muffintin effective potential"], state_ref["muffintin effective potential"])
    assert np.allclose(state_out["interstitial effective potential"], state_ref["interstitial effective potential"])
    assert np.allclose(state_out["reciprocal interstitial effective potential"],
                       state_ref["reciprocal interstitial effective potential"])



"""
Test parsers in simple_parser
"""

from excitingtools.parser_utils.simple_parser import match_current_return_line_n, \
    match_current_extract_from_line_n


def test_match_current_return_next_n():

    input_string = """
       Tolerance factor to reduce the MB size based on
       the eigenvectors of the bare Coulomb potential:   0.100000000000000     
     
    --------------------------------------------------------------------------------
     
     Screened Coulomb potential:
       Full-frequency Random-Phase Approximation
       Some additional data : 5
     """

    string_match = match_current_return_line_n(
        input_string, 'Screened Coulomb potential:', n_line=1
        )
    assert string_match.strip() == "Full-frequency Random-Phase Approximation", \
        "Expect to return the string one line below the match"

    string_match = match_current_return_line_n(input_string, 'Screened Coulomb potential:')
    assert string_match.strip() == "Full-frequency Random-Phase Approximation", \
        "Expect default behaviour to return the string one line below the match"

    string_match = match_current_return_line_n(
        input_string, 'Screened Coulomb potential:', n_line=2
        )
    assert string_match.strip() == "Some additional data : 5", \
        "Expect to return the string two lines below the match"


def test_match_current_extract_from_line_n():

    input_string = """
       Tolerance factor to reduce the MB size based on
       the eigenvectors of the bare Coulomb potential:   0.100000000000000     

    --------------------------------------------------------------------------------

     Screened Coulomb potential:
       Full-frequency Random-Phase Approximation
       Some additional data : 5
     """

    # Extract 5 from "Some additional data : 5"
    key_extraction = {'Screened Coulomb potential:': lambda x: int(x.split(':')[-1])}

    data = match_current_extract_from_line_n(input_string, key_extraction, n_line=2)

    assert data == {'Screened Coulomb potential:': 5}, \
        "Expect to return the integer from 2 lines below the match key"



"""Base class for exciting input classes.
"""
import importlib
import warnings
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Union, List, Type, Iterator
from xml.etree import ElementTree

import numpy as np

from excitingtools.exciting_dict_parsers.input_parser import parse_element_xml
from excitingtools.utils import valid_attributes as all_valid_attributes
from excitingtools.utils.dict_utils import check_valid_keys
from excitingtools.utils.jobflow_utils import special_serialization_attrs
from excitingtools.utils.utils import flatten_list

path_type = Union[str, Path]


class AbstractExcitingInput(ABC):
    """Base class for exciting inputs."""

    @property
    @abstractmethod
    def name(self) -> str:
        """ Tag of the xml subelement. """
        ...

    @abstractmethod
    def to_xml(self) -> ElementTree:
        """ Convert class attributes to XML ElementTree."""
        ...


class ExcitingXMLInput(AbstractExcitingInput, ABC):
    """Base class for exciting inputs, with exceptions being title, plan and qpointset,
     because they are not passed as a dictionary. """

    # Convert python data to string, formatted specifically for
    _attributes_to_input_str = {int: lambda x: str(x),
                                np.int64: lambda x: str(x),
                                np.float64: lambda x: str(x),
                                float: lambda x: str(x),
                                bool: lambda x: str(x).lower(),
                                str: lambda x: x,
                                list: lambda mylist: " ".join(str(x).lower() for x in mylist).strip(),
                                tuple: lambda mylist: " ".join(str(x).lower() for x in mylist).strip()
                                }

    def __init__(self, **kwargs):
        """Initialise class attributes with kwargs.

        Rather than define all options for a given method, pass as kwargs and directly
        insert as class attributes.

        Valid attributes, subtrees and mandatory attributes are taken automatically from
        the parsed schema, see [valid_attributes.py](excitingtools/utils/valid_attributes.py).
        """
        valid_attributes, valid_subtrees, mandatory_keys, multiple_children = self.get_valid_attributes()

        # check the keys
        missing_mandatory_keys = mandatory_keys - set(kwargs.keys())
        if missing_mandatory_keys:
            raise ValueError(f"Missing mandatory arguments: {missing_mandatory_keys}")
        check_valid_keys(kwargs.keys(), valid_attributes | set(valid_subtrees), self.name)

        # initialise the subtrees
        class_list = self._class_list_excitingtools()
        subtree_class_map = {cls.name: cls for cls in class_list}
        subtrees = set(kwargs.keys()) - valid_attributes
        single_subtrees = subtrees - multiple_children
        multiple_subtrees = subtrees - single_subtrees
        for subtree in single_subtrees:
            kwargs[subtree] = self._initialise_subelement_attribute(subtree_class_map[subtree], kwargs[subtree])
        for subtree in multiple_subtrees:
            kwargs[subtree] = [self._initialise_subelement_attribute(subtree_class_map[subtree], x) for
                               x in kwargs[subtree]]

        # Set attributes from kwargs
        self.__dict__.update(kwargs)

    def __setattr__(self, name: str, value):
        """ Overload the attribute setting in python with instance.attr = value to check for validity in the schema.

        :param name: name of the attribute
        :param value: new value, can be anything
        """
        valid_attributes, valid_subtrees, _, _ = self.get_valid_attributes()
        check_valid_keys({name}, valid_attributes | set(valid_subtrees), self.name)
        super().__setattr__(name, value)

    def __delattr__(self, name: str):
        mandatory_keys = list(self.get_valid_attributes())[2]
        if name in mandatory_keys:
            warnings.warn(f"Attempt to delete mandatory attribute '{name}' was prevented.")
        else:
            super().__delattr__(name)

    def get_valid_attributes(self) -> Iterator:
        """ Extract the valid attributes, valid subtrees, mandatory attributes and multiple children
        from the parsed schema.

        :return: valid attributes, valid subtrees, mandatory attributes and multiple children
        """
        yield set(getattr(all_valid_attributes, f"{self.name}_valid_attributes", set()))
        yield getattr(all_valid_attributes, f"{self.name}_valid_subtrees", [])
        yield set(getattr(all_valid_attributes, f"{self.name}_mandatory_attributes", set()))
        yield set(getattr(all_valid_attributes, f"{self.name}_multiple_children", set()))

    @staticmethod
    def _class_list_excitingtools() -> List[Type[AbstractExcitingInput]]:
        """ Find all exciting input classes in own module and excitingtools.
        """
        excitingtools_namespace_content = importlib.import_module("excitingtools").__dict__
        input_class_namespace_content = importlib.import_module("excitingtools.input.input_classes").__dict__
        all_contents = {**excitingtools_namespace_content, **input_class_namespace_content}.values()
        return [cls for cls in all_contents if isinstance(cls, type) and issubclass(cls, AbstractExcitingInput)]

    @staticmethod
    def _initialise_subelement_attribute(XMLClass, element):
        """ Initialize given elements to the ExcitingXSInput constructor. If element is already ExcitingXMLInput class
        object, nothing happens. Else the class constructor of the given XMLClass is called. For a passed
        dictionary the dictionary is passed as kwargs.
        """
        if isinstance(element, XMLClass):
            return element
        elif isinstance(element, dict):
            # assume kwargs
            return XMLClass(**element)
        else:
            # Assume the element type is valid for the class constructor
            return XMLClass(element)

    def to_xml(self) -> ElementTree:
        """Put class attributes into an XML tree, with the element given by self.name.

        Example ground state XML subtree:
           <groundstate vkloff="0.5  0.5  0.5" ngridk="2 2 2" mixer="msec" </groundstate>

        Note, kwargs preserve the order of the arguments, however the order does not appear to be
        preserved when passed to (or perhaps converted to string) with xml.etree.ElementTree.tostring.

        :return ElementTree.Element sub_tree: sub_tree element tree, with class attributes inserted.
        """
        valid_attributes, valid_subtrees, _, multiple_children = self.get_valid_attributes()

        attributes = {key: self._attributes_to_input_str[type(value)](value) for key, value
                      in vars(self).items() if key in valid_attributes}
        xml_tree = ElementTree.Element(self.name, **attributes)

        subtrees = {key: self.__dict__[key] for key in set(vars(self).keys()) - set(attributes.keys())}
        ordered_subtrees = flatten_list([subtrees[x] for x in valid_subtrees if x in subtrees])
        for subtree in ordered_subtrees:
            xml_tree.append(subtree.to_xml())

        # Seems to want this operation on a separate line
        xml_tree.text = ' '

        return xml_tree

    def to_xml_str(self) -> str:
        """ Convert attributes to XML tree string. """
        return ElementTree.tostring(self.to_xml(), encoding='unicode', method='xml')

    def as_dict(self) -> dict:
        """ Convert attributes to dictionary. """
        serialise_attrs = special_serialization_attrs(self)
        return {**serialise_attrs, "xml_string": self.to_xml_str()}

    @classmethod
    def from_xml(cls, xml_string: path_type):
        """ Initialise class instance from XML-formatted string.

        Example Usage
        --------------
        xs_input = ExcitingXSInput.from_xml(xml_string)
        """
        return cls(**parse_element_xml(xml_string, tag=cls.name))

    @classmethod
    def from_dict(cls, d):
        """ Recreates class instance from dictionary. """
        return cls.from_xml(d["xml_string"])


def query_exciting_version(exciting_root: path_type) -> dict:
    """Query the exciting version
    Inspect version.inc, which is constructed at compile-time.

    Assumes version.inc has this structure:
     #define GITHASH "1a2087b0775a87059d53"
     #define GITHASH2 "5d01a5475a10f00d0ad7"
     #define COMPILERVERSION "GNU Fortran (MacPorts gcc9 9.3.0_4) 9.3.0"
     #define VERSIONFROMDATE /21,12,01/

    TODO(Fab) Issue 117. Parse major version.
     Would need to parse src/mod_misc.F90 and regex for "character(40) :: versionname = "
     Refactor whole routine to use regex.

    :param exciting_root: exciting root directory.
    :return version: Build and version details
    """
    if isinstance(exciting_root, str):
        exciting_root = Path(exciting_root)

    version_inc = exciting_root / 'src/version.inc'

    if not version_inc.exists():
        raise FileNotFoundError(f'{version_inc} cannot be found. '
                                f'This file generated when the code is built')

    with open(version_inc, 'r') as fid:
        all_lines = fid.readlines()

    git_hash_part1 = all_lines[0].split()[-1][1:-1]
    git_hash_part2 = all_lines[1].split()[-1][1:-1]
    compiler_parts = all_lines[2].split()[2:]
    compiler = " ".join(s for s in compiler_parts).strip()

    version = {'compiler': compiler[1:-1], 'git_hash': git_hash_part1 + git_hash_part2}
    return version



"""
Tests for functions in dict_utils.py
"""
import numpy as np

from excitingtools.utils.dict_utils import container_converter, serialise_dict_values, delete_nested_key


def test_convert_container():
    """Test container_converter function on string values in a dict."""
    input = {
        'a': '5.0',
        'b': ['1.0', '10.0'],
        'c': {
            'a1': '1.0'
        },
        'd': "blu",
        'e': [['1', '2.3'], '3'],
        'f': ['1', 'a']
    }
    expected = {
        'a': 5.0,
        'b': [1.0, 10.0],
        'c': {
            'a1': 1.0
        },
        'd': "blu",
        'e': [[1, 2.3], 3],
        'f': [1, 'a']
    }

    assert container_converter(input) == expected, (
        'String value/s failed to convert to numerical values')


def test_convert_container_no_strings():
    """Test container_converter where there are no string values in dict."""
    input = {
        'a': 5.0,
        'b': [1.0, 10.0],
        'c': {
            'a1': 1.0
        },
        'd': "blu",
        'e': [[1, 2.3], 3],
        'f': [1, 11.3]
    }

    assert container_converter(input) == input, (
        "Expect the converter to do nothing")


def test_convert_container_data_type():
    input = {
        'a': '5.0',
        'b': ['1.0', '10.0'],
        'c': {
            'a1': '1.0'
        },
        'd': "blu",
        'e': [['1', '2.3'], '3']
    }
    expected = {
        'a': 5.0,
        'b': [1.0, 10.0],
        'c': {
            'a1': 1.0
        },
        'd': "blu",
        'e': [[1, 2.3], 3]
    }

    output = container_converter(input)

    for elem1, elem2 in zip(output.values(), expected.values()):
        assert type(elem1) == type(elem2)
    for elem, elem2 in zip(output['c'].values(), expected['c'].values()):
        assert type(elem) == type(elem2)
    for elem, elem2 in zip(output['b'], expected['b']):
        assert type(elem) == type(elem2)
    for elem, elem2 in zip(output['e'], expected['e']):
        assert type(elem) == type(elem2)
    for elem, elem2 in zip(output['e'][0], expected['e'][0]):
        assert type(elem) == type(elem2)

    assert output == expected, (
        "Output is consistent with the expected dictionary")


class Mock:

    def __init__(self, a, b):
        self.a = a
        self.b = b


def test_serialise_dict_values():

    # Value is an object
    input = {'mock_key': Mock(1, 2)}
    output = serialise_dict_values(input)
    assert output == {
        'mock_key': {
            'a': 1,
            'b': 2
        }
    }, "Convert object values to dicts"

    # Object nested in a dictionary
    input = {'mock_key': {'another-key': Mock(1, 2)}}
    output = serialise_dict_values(input)
    assert output == {
        'mock_key': {
            'another-key': {
                'a': 1,
                'b': 2
            }
        }
    }, "Convert nested object values into dicts"

    # Object nested in a list, and within a dictionary within a list
    input = {'a': [1, 2, Mock(3, 4), {'b': Mock(5, 6)}]}
    output = serialise_dict_values(input)
    assert output == {'a': [1, 2, {'a': 3, 'b': 4}, {'b': {'a': 5, 'b': 6}}]}, \
        "Convert nested object values into dicts, where the top-level container value is a list"


def test_serialise_dict_value_is_tuple():

    # Object nested in a dictionary, nested within a tuple
    input = {'mock_key': (1, {'another-key': Mock(1, 2)})}
    output = serialise_dict_values(input)
    assert output == {'mock_key': [1, {'another-key': {'a': 1, 'b': 2}}]}, \
        "Convert nested object values into dicts, where the top-level container value is a tuple." \
        "Note, tuple cannot be mutated so it is converted to a list by serialise_dict_values"


def test_serialise_dict_value_is_set():

    # Object nested in a dictionary, nested within a set
    input = {'mock_key': {1, 2, Mock(1, 2)}}
    output = serialise_dict_values(input)

    output_keys = [k for k in output.keys()]
    output_values = [v for v in output.values()]

    assert len(output_keys) == 1
    assert output_keys[0] == 'mock_key'
    assert len(output_values) == 1

    # Mock object -> dictionary, meaning one cannot compare the input and outout with collections.Counter,
    # sets or sorted (neither object nor dictionary is hashable)
    def unordered_lists_the_same(a: list, b: list) -> bool:
        """
        Check contents of two lists are consistent, irregardless of order
        """
        for element in a:
            try:
                b.remove(element)
            except ValueError:
                return False
        return not b

    assert unordered_lists_the_same(output_values[0], [1, 2, {'a': 1, 'b': 2}]), \
        "Convert nested object values into dicts, where the top-level container value is a set. " \
        "Note, set is not iterable and is converted to a list with some arbitrary order by serialise_dict_values"


def test_serialise_dict_values_null_behaviour():
    """"
    Test serialize_dict_values on different dict key types.
    """
    input = {'mock_key': [1, 2]}
    output = serialise_dict_values(input)
    assert output == input, "Pass over list values"

    input = {'mock_key': (1, 2)}
    output = serialise_dict_values(input)
    assert output == input, "Pass over tuple values"

    input = {'mock_key': {1, 2}}
    output = serialise_dict_values(input)
    assert output == input, "Pass over set values"

    input = {'mock_key': np.array([1., 2., 3.])}
    output = serialise_dict_values(input)
    assert output == input, "Pass over np.array values"


def test_delete_nested_key():
    """
    Test delete_nested_key removes a key from dict.
    """
    input = {'a': {'b': 1, 'c': {'d': 1, 'e': 2}}}
    key_to_remove = ['a']
    delete_nested_key(input, key_to_remove)
    output = {}
    assert output == input

    input = {'a': {'b': 1, 'c': {'d': 1, 'e': 2}}}
    key_to_remove = ['a', 'c', 'd']
    delete_nested_key(input, key_to_remove)
    output = {'a': {'b': 1, 'c': {'e': 2}}}
    assert output == input



"""
File containing pytest fixtures. They can be seen from all other files and subdirectories.
That's because of the special name of the file.
Needed for testing environment variables.
"""
import pytest


@pytest.fixture
def mock_env_jobflow(monkeypatch):
    monkeypatch.setenv("USE_JOBFLOW", "true")


@pytest.fixture
def mock_env_jobflow_missing(monkeypatch):
    monkeypatch.delenv("USE_JOBFLOW", raising=False)



""" Test the schema parsing and whether the file valid_attributes is up-to-date. """

from excitingtools.utils.schema_parsing import list_string_line_limit, get_excitingtools_root


def test_set_string_line_limit():
    properties_valid_subtrees = ['DFTD2', 'EFG', 'LSJ', 'TSvdW', 'bandstructure', 'boltzequ', 'chargedensityplot',
                                 'momentummatrix', 'mossbauer', 'mvecfield', 'polarization', 'raman', 'shg', 'spintext',
                                 'stm', 'wannier', 'wanniergap', 'wannierplot', 'wfplot', 'xcmvecfield']
    reference_string = ("properties_valid_subtrees = ['DFTD2', 'EFG', 'LSJ', 'TSvdW', "
                        "'bandstructure', 'boltzequ', 'chargedensityplot', \n"
                        "                             'momentummatrix', 'mossbauer', 'mvecfield', "
                        "'polarization', 'raman', 'shg', 'spintext', \n"
                        "                             'stm', 'wannier', 'wanniergap', 'wannierplot', "
                        "'wfplot', 'xcmvecfield']")
    assert list_string_line_limit("properties_valid_subtrees", properties_valid_subtrees) == reference_string


def test_get_exciting_root():
    assert get_excitingtools_root().name == "exciting_tools"



"""
Test for the input.xml file parser
"""
import pytest

from excitingtools.exciting_dict_parsers.input_parser import parse_element_xml, parse_structure, parse_input_xml

reference_input_str = """<?xml version="1.0" encoding="UTF-8"?>
<input sharedfs="true" 
  xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" 
  xsi:noNamespaceSchemaLocation="https://xml.exciting-code.org/excitinginput.xsd">
  
  <title>Lithium Fluoride BSE</title>
  
  <structure speciespath="." autormt="false" epslat="1.0d-6">
    <crystal scale="1.0" stretch="1.0 1.0 1.0">
      <basevect>3.80402 3.80402 0.00000</basevect>
      <basevect>3.80402 0.00000 3.80402</basevect>
      <basevect>0.00000 3.80402 3.80402</basevect>
    </crystal>
    <species speciesfile="Li.xml" rmt="1.5">
      <atom coord="0.0000  0.0000  0.0000" bfcmt="0.0 0.0 0.0"/>
      <dfthalfparam 
        cut="3.90" 
        ampl="1" 
        exponent="8">
        <shell number="0" ionization="0.25" />
      </dfthalfparam>
    </species>
    <species speciesfile="F.xml">
      <atom coord="0.5000  0.5000  0.5000" lockxyz="false true false"/>
      <LDAplusU J="2.3" U="0.5" l="3"/>
    </species>
  </structure>
  
  <groundstate xctype="GGA_PBE" ngridk="4 4 4" epsengy="1d-7" outputlevel="high">
    <spin bfieldc="0 0 0" fixspin="total FSM"/>
    <OEP maxitoep="100"> </OEP>
  </groundstate>
  
  <properties>
    <dos 
      nsmdos="2"
      ngrdos="300"
      nwdos="1000"
      winddos="-0.3 0.3">
    </dos>
    <bandstructure>
      <plot1d>
        <path steps="100">
          <point coord="1.0     0.0     0.0" label="Gamma"/>
          <point coord="0.625   0.375   0.0" label="K"/>
          <point coord="0.5     0.5     0.0" label="X" breakafter="true"/>
          <point coord="0.0     0.0     0.0" label="Gamma"/>
          <point coord="0.5     0.0     0.0" label="L"/>
        </path>
      </plot1d>
    </bandstructure>
  </properties>

  <xs xstype="BSE" 
   ngridq="3 3 3"
   vkloff="0.05 0.15 0.25" 
   nempty="1"
   broad="0.0073499"
   nosym="true">

   <energywindow intv="0.0 1.0" 
    points="50" />

   <screening screentype="full"
    nempty="115" />

   <BSE bsetype="singlet"
    nstlbse="1 5 1 2" 
    aresbse="false"/>

   <qpointset>
      <qpoint>0.0 0.0 0.0</qpoint>
   </qpointset>
   
   <plan>
      <doonly task="screen" />
      <doonly task="bse" />
   </plan>
  </xs>
  
</input>
"""


def test_parse_title():
    assert parse_element_xml(reference_input_str, tag="title") == "Lithium Fluoride BSE"


def test_parse_groundstate():
    ground_state = parse_element_xml(reference_input_str, tag="groundstate")
    assert ground_state == {
        'xctype': 'GGA_PBE', 'ngridk': [4, 4, 4],
        'epsengy': 1e-7, 'outputlevel': 'high',
        'spin': {'bfieldc': [0, 0, 0], 'fixspin': 'total FSM'},
        'OEP': {'maxitoep': 100}
    }


def test_parse_groundstate_from_gs_root():
    ground_state = parse_element_xml('<groundstate xctype="GGA_PBE" ngridk="4 4 4" epsengy="1d-7" outputlevel="high"/>',
                                     tag="groundstate")
    assert ground_state == {
        'xctype': 'GGA_PBE', 'ngridk': [4, 4, 4],
        'epsengy': 1e-7, 'outputlevel': 'high'
    }


def test_parse_structure():
    structure = parse_structure(reference_input_str)
    structure_ref = {
        'atoms': [{'species': 'Li', 'position': [0.0, 0.0, 0.0],
                   'bfcmt': [0.0, 0.0, 0.0]},
                  {'species': 'F', 'position': [0.5, 0.5, 0.5],
                   'lockxyz': [False, True, False]}],
        'lattice': [[3.80402, 3.80402, 0.0],
                    [3.80402, 0.0, 3.80402],
                    [0.0, 3.80402, 3.80402]],
        'species_path': '.',
        'crystal_properties': {'scale': 1.0, 'stretch': [1.0, 1.0, 1.0]},
        'species_properties': {'F': {'LDAplusU': {'J': 2.3, 'U': 0.5, 'l': 3}},
                               'Li': {'dfthalfparam': {'ampl': 1,
                                                       'cut': 3.9,
                                                       'exponent': 8,
                                                       'shell': [{'ionization': 0.25,
                                                                 'number': 0}]},
                                      'rmt': 1.5}},
        'autormt': False,
        'epslat': 1.0e-6,
    }
    assert structure_ref == structure


def test_parse_xs():
    xs = parse_element_xml(reference_input_str, tag="xs")
    xs_ref = {
        'xstype': 'BSE',
        'ngridq': [3, 3, 3],
        'vkloff': [0.05, 0.15, 0.25],
        'nempty': 1,
        'broad': 0.0073499,
        'nosym': True,
        'energywindow': {'intv': [0.0, 1.0], 'points': 50},
        'screening': {'screentype': 'full', 'nempty': 115},
        'BSE': {'bsetype': 'singlet', 'nstlbse': [1, 5, 1, 2], 'aresbse': False},
        'qpointset': [[0.0, 0.0, 0.0]],
        'plan': ['screen', 'bse']
    }
    assert xs_ref == xs
    assert isinstance(xs["ngridq"][0], int)


input_ref_parsed_keys = {'title', 'groundstate', 'structure', 'xs', 'sharedfs', "properties"}


def test_parse_input_xml():
    parsed_data = parse_element_xml(reference_input_str)
    assert set(parsed_data.keys()) == input_ref_parsed_keys
    assert parsed_data['sharedfs']


def test_parse_input_xml_directly():
    parsed_data = parse_input_xml(reference_input_str)
    assert set(parsed_data.keys()) == input_ref_parsed_keys


def test_parse_missing_tag():
    with pytest.raises(ValueError, match="Your specified input has no tag missing_tag"):
        parse_element_xml(reference_input_str, "missing_tag")


def test_parse_input_xml_with_tag():
    parsed_data = parse_element_xml(reference_input_str, tag="input")
    assert set(parsed_data.keys()) == input_ref_parsed_keys


def test_parse_properties():
    properties = parse_element_xml(reference_input_str, tag="properties")
    properties_ref = {
        'dos': {"nsmdos": 2, "ngrdos": 300, "nwdos": 1000, "winddos": [-0.3, 0.3]},
        'bandstructure': {"plot1d": {"path": {
            "steps": 100,
            "point":
                [
                    {"coord": [1, 0, 0], "label": "Gamma"},
                    {"coord": [0.625, 0.375, 0], "label": "K"},
                    {"coord": [0.5, 0.5, 0], "label": "X", "breakafter": True},
                    {"coord": [0, 0, 0], "label": "Gamma"},
                    {"coord": [0.5, 0, 0], "label": "L"}
                ]}}}
    }
    assert properties_ref == properties



""" Units.

Physical constants, defined according to [CODATA 2018])(http://physics.nist.gov/constants)
One could also consider importing them from scipy
"""
import enum

bohr_to_angstrom = 0.529177210903
angstrom_to_bohr = 1. / bohr_to_angstrom
Hartree_to_eV = 27.211396641308


class Unit(enum.Enum):
    """
    Enum class for exciting units. All names are defined with
    as lowercase, for consistency.

    This could be replaced with [PINT](https://pint.readthedocs.io/en/stable/), as used by NOMAD, however it's
    currently not required. If/when we wish to do unit manipulation, it should be reconsidered.
    """
    hartree = enum.auto()
    inv_hartree = enum.auto()
    ev = enum.auto()
    inv_ev = enum.auto()
    kelvin = enum.auto()
    bohr = enum.auto()
    bohr_pow_3 = enum.auto()
    inv_bohr = enum.auto()
    inv_bohr_pow_3 = enum.auto()
    au = enum.auto()
    degrees = enum.auto()
    GK_max = enum.auto()
    electron_rest_mass = enum.auto()
    bohr_velocity_over_bohr_radius = enum.auto()
    bohr_velocity = enum.auto()
    force = enum.auto()
    null = enum.auto()


# Map Unit enums to strings.
# Required because JSON cannot dump objects to file.
enum_to_string = {
    Unit.hartree: 'Hartree',
    Unit.inv_hartree: '1/Hartree',
    Unit.ev: 'eV',
    Unit.inv_ev: 'eV^-1',
    Unit.kelvin: 'K',
    Unit.bohr: 'Bohr',
    Unit.bohr_pow_3: 'Bohr^3',
    Unit.inv_bohr: 'Bohr^-1',
    Unit.inv_bohr_pow_3: 'Bohr^-3',
    Unit.au: 'a.u.',
    Unit.degrees: 'degrees',
    Unit.GK_max: 'GK_max',
    Unit.electron_rest_mass: 'm_electron',
    Unit.bohr_velocity_over_bohr_radius: 'v_Bohr/r_Bohr',
    Unit.bohr_velocity: 'Bohr/t_Bohr',
    Unit.force: 'Hartree/Bohr', 
    Unit.null: 'null'
}



""" Automatic generation of all standard input classes plus definiton of exceptions. """

from pathlib import Path
from typing import Optional, List, Callable
from typing import Union
from xml.etree import ElementTree

import numpy as np

from excitingtools.input.base_class import AbstractExcitingInput, ExcitingXMLInput
from excitingtools.input.dynamic_class import generate_classes_str
from excitingtools.input.xml_utils import xml_tree_to_pretty_str, prettify_tag_attributes
from excitingtools.utils.dict_utils import check_valid_keys
from excitingtools.utils.utils import list_to_str
from excitingtools.utils.valid_attributes import valid_plan_entries

# define names of classes which are meant to be available for a user or used directly elsewhere in excitingtools
ExcitingCrystalInput: Callable
ExcitingSpeciesInput: Callable
ExcitingGroundStateInput: Callable
ExcitingXSInput: Callable
ExcitingPropertiesInput: Callable
ExcitingPointInput: Callable
ExcitingBandStructureInput: Callable

# execute dynamically generated string with all standard class defintions
exec(generate_classes_str())


class ExcitingTitleInput(AbstractExcitingInput):
    """ Holds only the title but for consistency reasons as class. """
    name = "title"

    def __init__(self, title: str):
        self.title = title

    def to_xml(self, **kwargs) -> ElementTree:
        """ Puts title to xml, only the text is title. """
        title_tree = ElementTree.Element(self.name)
        title_tree.text = self.title
        return title_tree


class ExcitingInputXML(ExcitingXMLInput):
    """
    Container for a complete input xml file.
    """
    name = "input"
    _default_filename = "input.xml"

    def set_title(self, title: str):
        """ Set a new title. """
        self.__dict__["title"].title = title

    def to_xml_str(self) -> str:
        """Compose XML ElementTrees from exciting input classes to create an input xml string.

        :return: Input XML tree as a string, with pretty formatting.
        """
        return prettify_tag_attributes(xml_tree_to_pretty_str(self.to_xml()))

    def write(self, filename: Union[str, Path] = _default_filename):
        """Writes the xml string to file.

        :param filename: name of the file.
        """
        with open(filename, "w") as fid:
            fid.write(self.to_xml_str())


class ExcitingQpointsetInput(AbstractExcitingInput):
    """
    Class for exciting Qpointset Input
    """
    name = "qpointset"

    def __init__(self, qpointset: Optional[Union[np.ndarray, List[List[float]]]] = np.array([0.0, 0.0, 0.0])):
        """
        Qpointset should be passed either as numpy array or as a list of lists, so either
        np.array([[0., 0., 0.], [0.0, 0.0, 0.01], ...])
        or
        [[0., 0., 0.], [0.0, 0.0, 0.01], ...]
        """
        self.qpointset = qpointset

    def to_xml(self) -> ElementTree.Element:
        """ Special implementation of to_xml for the qpointset element. """
        qpointset = ElementTree.Element(self.name)
        for qpoint in self.qpointset:
            ElementTree.SubElement(qpointset, 'qpoint').text = list_to_str(qpoint)

        return qpointset


class ExcitingPlanInput(AbstractExcitingInput):
    """
    Class for exciting Plan Input
    """
    name = "plan"

    def __init__(self, plan: List[str]):
        """
        Plan doonly elements are passed as a List of strings in the order exciting shall execute them:
            ['bse', 'xseigval', ...]
        """
        check_valid_keys(plan, valid_plan_entries, self.name)
        self.plan = plan

    def to_xml(self) -> ElementTree.Element:
        """ Special implementation of to_xml for the plan element. """
        plan = ElementTree.Element(self.name)
        for task in self.plan:
            ElementTree.SubElement(plan, 'doonly', task=task)

        return plan



"""Test ExcitingStructure, python API that generates exciting's structure XML.

NOTE:
All attribute tests should assert on the XML tree content, as the attribute
order is not preserved by the ElementTree.tostring method. Elements appear to
be fine.

For example:

 `gs_xml_string = xml.etree.ElementTree.tostring(
     gs_xml, encoding='unicode', method='xml')`

may return:

'<groundstate ngridk="8 8 8" rgkmax="8.6"> </groundstate>'
or
'<groundstate rgkmax="8.6" ngridk="8 8 8"> </groundstate>'

"""

import numpy as np
import pytest

from excitingtools.input.bandstructure import get_bandstructure_input_from_exciting_structure
from excitingtools.input.structure import ExcitingStructure


@pytest.fixture
def xml_structure_H2He():
    """
    structure object initialised with a mock crystal, using mandatory arguments only.
    """
    cubic_lattice = [[1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]]
    arbitrary_atoms = [{'species': 'H', 'position': [0, 0, 0]},
                       {'species': 'H', 'position': [1, 0, 0]},
                       {'species': 'He', 'position': [2, 0, 0]}]
    structure = ExcitingStructure(arbitrary_atoms, cubic_lattice, './')
    return structure.to_xml()


def test_class_exciting_structure_xml(xml_structure_H2He):
    """
    Test input XML attributes from an instance of ExcitingStructure.
    """
    assert xml_structure_H2He.tag == 'structure', 'XML root should be structure'
    assert xml_structure_H2He.keys() == ['speciespath'], 'structure defined to have only speciespath '
    assert xml_structure_H2He.get('speciespath') == './', 'species path set to ./'


def test_class_exciting_structure_crystal_xml(xml_structure_H2He):
    """
    crystal subtree of structure.
    """
    elements = list(xml_structure_H2He)
    assert len(elements) == 3, 'Expect structure tree to have 3 sub-elements'

    crystal_xml = elements[0]
    assert crystal_xml.tag == "crystal", 'First subtree is crystal'
    assert crystal_xml.items() == [], "No required attributes in crystal."

    lattice_vectors = list(crystal_xml)
    assert len(lattice_vectors) == 3, "Always expect three lattice vectors"
    assert lattice_vectors[0].items() == [], 'Lattice vectors have no items'
    assert lattice_vectors[0].text == "1.0 0.0 0.0", "Lattice vector `a` differs from input"
    assert lattice_vectors[1].text == "0.0 1.0 0.0", "Lattice vector `b` differs from input"
    assert lattice_vectors[2].text == "0.0 0.0 1.0", "Lattice vector `c` differs from input"


def test_class_exciting_structure_species_xml(xml_structure_H2He):
    """
    species subtrees of structure.
    """
    elements = list(xml_structure_H2He)
    assert len(elements) == 3, 'Expect structure tree to have 3 sub-elements'

    species_h_xml = elements[1]
    assert species_h_xml.tag == "species", 'Second subtree is species'

    species_he_xml = elements[2]
    assert species_he_xml.tag == "species", 'Third subtree is species'

    assert species_h_xml.items() == [('speciesfile', 'H.xml')], 'species is inconsistent'
    assert species_he_xml.items() == [('speciesfile', 'He.xml')], 'species is inconsistent'

    atoms_h = list(species_h_xml)
    assert len(atoms_h) == 2, 'Number of H atoms is wrong'
    assert atoms_h[0].items() == [('coord', '0 0 0')], "Coordinate of first H differs to input"
    assert atoms_h[1].items() == [('coord', '1 0 0')], "Coordinate of second H differs to input"

    atoms_he = list(species_he_xml)
    assert len(atoms_he) == 1, 'Number of He atoms is wrong'
    assert atoms_he[0].items() == [('coord', '2 0 0')], "Coordinate of only He differs to input"


@pytest.fixture
def xml_structure_CdS():
    """
    structure object initialised with a mock crystal, using all atom properties.
    Optional atom attributes require the generic container, List[dict].
    """
    cubic_lattice = [[1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]]
    arbitrary_atoms = [{
        'species': 'Cd',
        'position': [0.0, 0.0, 0.0],
        'bfcmt': [1.0, 1.0, 1.0],
        'lockxyz': [False, False, False],
        'mommtfix': [2.0, 2.0, 2.0]
    }, {'species': 'S', 'position': [1.0, 0.0, 0.0]}]
    structure = ExcitingStructure(arbitrary_atoms, cubic_lattice, './')
    return structure.to_xml()


def test_optional_atom_attributes_xml(xml_structure_CdS):
    """
    Test optional atom attributes are correctly set in XML tree.
    """
    assert xml_structure_CdS.tag == 'structure'
    assert xml_structure_CdS.keys() == ['speciespath'], 'structure defined to have only speciespath '
    assert xml_structure_CdS.get('speciespath') == './', 'speciespath set to be ./'

    elements = list(xml_structure_CdS)
    assert len(elements) == 3, 'Expect structure tree to have 3 sub-elements'

    # Crystal
    crystal_xml = elements[0]
    assert crystal_xml.tag == "crystal", 'First subtree is crystal'
    assert crystal_xml.items() == [], "No required attributes in crystal."

    # Species
    species_cd_xml = elements[1]
    assert species_cd_xml.tag == "species", 'Second subtree is species'
    assert species_cd_xml.items() == [('speciesfile', 'Cd.xml')]

    species_s_xml = elements[2]
    assert species_s_xml.tag == "species", 'Third subtree is species'
    assert species_s_xml.items() == [('speciesfile', 'S.xml')]

    # Cd
    atoms_cd = list(species_cd_xml)
    assert len(atoms_cd) == 1, 'Wrong number of Cd atoms'
    assert set(atoms_cd[0].keys()) == {'coord', 'bfcmt', 'mommtfix', 'lockxyz'}, \
        'Cd contains all mandatory and optional atom properties'
    assert ('coord', '0.0 0.0 0.0') in atoms_cd[0].items()
    assert ('bfcmt', '1.0 1.0 1.0') in atoms_cd[0].items()
    assert ('mommtfix', '2.0 2.0 2.0') in atoms_cd[0].items()
    assert ('lockxyz', 'false false false') in atoms_cd[0].items()

    # S
    atoms_s = list(species_s_xml)
    assert len(atoms_s) == 1, 'Wrong number of S atoms'
    assert atoms_s[0].keys() == ['coord'], \
        'S only contains mandatory atom properties'
    assert atoms_s[0].items() == [('coord', '1.0 0.0 0.0')]


@pytest.fixture
def lattice_and_atoms_CdS():
    """
    structure object initialised with a mock crystal
    """
    cubic_lattice = [[1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]]
    arbitrary_atoms = [{'species': 'Cd', 'position': [0.0, 0.0, 0.0]},
                       {'species': 'S', 'position': [1.0, 0.0, 0.0]}]
    return cubic_lattice, arbitrary_atoms


def test_optional_structure_attributes_xml(lattice_and_atoms_CdS):
    """
    Test optional structure attributes.
    """
    cubic_lattice, arbitrary_atoms = lattice_and_atoms_CdS
    structure_attributes = {
        'autormt': True, 'cartesian': False, 'epslat': 1.e-6, 'primcell': False, 'tshift': True
    }
    structure = ExcitingStructure(arbitrary_atoms, cubic_lattice, './', **structure_attributes)
    xml_structure = structure.to_xml()

    mandatory = {'speciespath'}
    optional = set(structure_attributes)

    assert xml_structure.tag == 'structure'
    assert set(xml_structure.keys()) == mandatory | optional, \
        'Should contain mandatory speciespath plus all optional attributes'
    assert xml_structure.get('speciespath') == './', 'species path should be ./'
    assert xml_structure.get('autormt') == 'true'
    assert xml_structure.get('cartesian') == 'false'
    assert xml_structure.get('epslat') == '1e-06'
    assert xml_structure.get('primcell') == 'false'
    assert xml_structure.get('tshift') == 'true'


def test_optional_crystal_attributes_xml(lattice_and_atoms_CdS):
    """
    Test optional crystal attributes.
    """
    cubic_lattice, arbitrary_atoms = lattice_and_atoms_CdS

    structure = ExcitingStructure(
        arbitrary_atoms,
        cubic_lattice,
        './',
        crystal_properties={'scale': 1.00, 'stretch': [1.00, 1.00, 1.00]}
    )
    xml_structure = structure.to_xml()

    elements = list(xml_structure)
    assert len(elements) == 3, 'Number of sub-elements in structure tree'

    crystal_xml = elements[0]
    assert crystal_xml.tag == "crystal", 'First subtree is crystal'
    assert crystal_xml.keys() == ['scale', 'stretch'], 'Optional crystal properties'
    assert crystal_xml.get('scale') == '1.0', 'scale value inconsistent with input'
    assert crystal_xml.get('stretch') == '1.0 1.0 1.0', 'stretch value inconsistent with input'


def test_optional_species_attributes_xml(lattice_and_atoms_CdS):
    """
    Test optional species attributes.
    """
    cubic_lattice, arbitrary_atoms = lattice_and_atoms_CdS
    species_attributes = {'Cd': {'rmt': 3.0, "LDAplusU": {"J": 1.5, "U": 2.4, "l": 2}},
                          'S': {'rmt': 4.0, "dfthalfparam": {"ampl": 1.2, "cut": 1.9, "exponent": 5,
                                                             "shell": [{"ionization": 0.8, "number": 1}]}}}

    structure = ExcitingStructure(
        arbitrary_atoms, cubic_lattice, './', species_properties=species_attributes
    )
    xml_structure = structure.to_xml()

    elements = list(xml_structure)
    assert len(elements) == 3, 'Number of sub-elements in structure tree'

    species_cd_xml = elements[1]
    assert species_cd_xml.tag == "species", 'Second subtree is species'

    species_s_xml = elements[2]
    assert species_s_xml.tag == "species", 'Third subtree is species'

    assert set(species_cd_xml.keys()) == {'speciesfile', 'rmt'}, "species attributes differ from expected"
    assert species_cd_xml.get('speciesfile') == 'Cd.xml', 'speciesfile differs from expected'
    assert species_cd_xml.get('rmt') == '3.0', 'Cd muffin tin radius differs from input'

    species_cd_elements = list(species_cd_xml)
    assert len(species_cd_elements) == 2
    ldaplusu_xml = species_cd_elements[0]
    assert ldaplusu_xml.tag == "LDAplusU"

    assert set(ldaplusu_xml.keys()) == {'J', 'U', 'l'}
    assert ldaplusu_xml.get("J") == "1.5"
    assert ldaplusu_xml.get("U") == "2.4"
    assert ldaplusu_xml.get("l") == "2"

    assert set(species_s_xml.keys()) == {'speciesfile', 'rmt'}, "species attributes differ from expected"
    assert species_s_xml.get('speciesfile') == 'S.xml', 'speciesfile differs from expected'
    assert species_s_xml.get('rmt') == '4.0', 'S muffin tin radius differs from input'

    species_s_elements = list(species_s_xml)
    assert len(species_s_elements) == 2
    dfthalfparam_xml = species_s_elements[0]
    assert dfthalfparam_xml.tag == "dfthalfparam"

    assert set(dfthalfparam_xml.keys()) == {'ampl', 'cut', 'exponent'}
    assert dfthalfparam_xml.get("ampl") == "1.2"
    assert dfthalfparam_xml.get("cut") == "1.9"
    assert dfthalfparam_xml.get("exponent") == "5"

    dfthalfparam_elements = list(dfthalfparam_xml)
    assert len(dfthalfparam_elements) == 1
    shell_xml = dfthalfparam_elements[0]
    assert shell_xml.tag == "shell"

    assert set(shell_xml.keys()) == {'ionization', 'number'}
    assert shell_xml.get('ionization') == '0.8'
    assert shell_xml.get('number') == '1'


ref_dict = {'xml_string': '<structure speciespath="./"> <crystal> <basevect>1.0 0.0 0.0</basevect>'
                          '<basevect>0.0 1.0 0.0</basevect><basevect>0.0 0.0 1.0</basevect></crystal>'
                          '<species speciesfile="Cd.xml"> <atom coord="0.0 0.0 0.0"> </atom></species>'
                          '<species speciesfile="S.xml"> <atom coord="1.0 0.0 0.0"> </atom></species></structure>'}


def test_as_dict(lattice_and_atoms_CdS, mock_env_jobflow_missing):
    cubic_lattice, arbitrary_atoms = lattice_and_atoms_CdS

    structure = ExcitingStructure(arbitrary_atoms, cubic_lattice, './')

    assert structure.as_dict() == ref_dict, 'expected different dict representation'


def test_as_dict_jobflow(lattice_and_atoms_CdS, mock_env_jobflow):
    cubic_lattice, arbitrary_atoms = lattice_and_atoms_CdS

    structure = ExcitingStructure(arbitrary_atoms, cubic_lattice, './')

    assert structure.as_dict() == {**ref_dict, '@class': 'ExcitingStructure',
                                   '@module': 'excitingtools.input.structure'}, \
        'expected different dict representation'


def test_from_dict(lattice_and_atoms_CdS):
    cubic_lattice, arbitrary_atoms = lattice_and_atoms_CdS
    structure = ExcitingStructure.from_dict(ref_dict)

    assert np.allclose(structure.lattice, np.array(cubic_lattice))
    assert structure.species == [d['species'] for d in arbitrary_atoms]
    assert structure.positions == [d['position'] for d in arbitrary_atoms]
    assert structure.speciespath == './'  # pylint: disable=no-member


@pytest.fixture
def lattice_and_atoms_H20():
    """
    H20 molecule in a big box (angstrom)
    """
    cubic_lattice = [[10.0, 0.0, 0.0], [0.0, 10.0, 0.0], [0.0, 0.0, 10.0]]
    atoms = [{'species': 'H', 'position': [0.00000, 0.75545, -0.47116]},
             {'species': 'O', 'position': [0.00000, 0.00000, 0.11779]},
             {'species': 'H', 'position': [0.00000, 0.75545, -0.47116]}]
    return cubic_lattice, atoms


def test_group_atoms_by_species(lattice_and_atoms_H20):
    """
    Test group_atoms_by_species method.
    """
    cubic_lattice, atoms = lattice_and_atoms_H20
    structure = ExcitingStructure(atoms, cubic_lattice, './')
    assert structure.species == ['H', 'O', 'H'], 'Species list differs from lattice_and_atoms_H20'

    indices = dict(structure._group_atoms_by_species())
    assert set(indices.keys()) == {'H', 'O'}, 'List unique species in structure'
    assert indices['H'] == [0, 2], "Expect atoms 0 and 2 to be H"
    assert indices['O'] == [1], "Expect atom 1 to be O"

    hydrogen = [structure.species[i] for i in indices['H']]
    oxygen = [structure.species[i] for i in indices['O']]
    assert hydrogen == ["H", "H"], "Expect list to only contain H symbols"
    assert oxygen == ["O"], "Expect list to contain only one O symbol"


@pytest.fixture
def ase_atoms_H20(lattice_and_atoms_H20):
    """
    H20 molecule in a big box (angstrom), in ASE Atoms()
    Converts a List[dict] to ase.atoms.Atoms.
    """
    ase = pytest.importorskip("ase")
    lattice, atoms = lattice_and_atoms_H20
    symbols = [atom['species'] for atom in atoms]
    cubic_cell = np.asarray(lattice)
    positions = [atom['position'] for atom in atoms]
    return ase.Atoms(symbols=symbols, positions=positions, cell=cubic_cell, pbc=True)


def test_class_exciting_structure_ase(ase_atoms_H20):
    """
    Test the ASE Atoms object gets used correctly by the ExcitingStructure constructor.
    """
    atoms = ase_atoms_H20
    structure = ExcitingStructure(atoms, species_path='./')

    assert structure.species == ["H", "O", "H"]
    assert np.allclose(structure.lattice,
                       [[18.897261246257703, 0.0, 0.0],
                        [0.0, 18.897261246257703, 0.0],
                        [0.0, 0.0, 18.897261246257703]]), \
        'Expect lattice vectors to match input values'

    assert np.allclose(structure.positions, atoms.get_scaled_positions()), 'Expect positions to match input values.'

    # TODO(Alex) Issue 117. Compare xml_structure built with and without ASE - should be consistent
    # This just confirms the XML tree is built, not that it is correct.
    xml_structure = structure.to_xml()
    assert list(xml_structure.keys()) == ['speciespath'], 'Only expect speciespath in structure xml keys'


def test_using_non_standard_species_symbol():
    cubic_lattice = [[10.0, 0.0, 0.0], [0.0, 10.0, 0.0], [0.0, 0.0, 10.0]]
    atoms = [{'species': 'C_molecule', 'position': [0.00000, 0.75545, -0.47116]}]
    structure = ExcitingStructure(atoms, cubic_lattice)

    structure_xml = structure.to_xml()
    assert structure_xml.tag == 'structure', 'XML root should be structure'
    assert structure_xml.keys() == ['speciespath'], 'structure defined to have only speciespath '
    assert structure_xml.get('speciespath') == './', 'species path set to ./'

    elements = list(structure_xml)
    assert len(elements) == 2, 'Expect structure tree to have 2 sub-elements'

    species_c_xml = elements[1]
    assert species_c_xml.tag == "species", 'Second subtree is species'

    assert species_c_xml.items() == [('speciesfile', 'C_molecule.xml')], 'species is inconsistent'

    atoms_h = list(species_c_xml)
    assert len(atoms_h) == 1
    assert atoms_h[0].items() == [('coord', '0.0 0.75545 -0.47116')]


def test_get_full_lattice(lattice_and_atoms_CdS):
    cubic_lattice, arbitrary_atoms = lattice_and_atoms_CdS
    structure = ExcitingStructure(arbitrary_atoms, cubic_lattice, './',
                                  crystal_properties={'scale': 1.50, 'stretch': [2.00, 1.00, 3.00]})
    ref_lattice = np.array([[3, 0, 0], [0, 1.5, 0], [0, 0, 4.5]])
    assert np.allclose(structure.get_lattice(), ref_lattice)


def test_get_bandstructure_input_from_exciting_structure(lattice_and_atoms_H20):
    pytest.importorskip("ase")
    cubic_lattice, atoms = lattice_and_atoms_H20
    structure = ExcitingStructure(atoms, cubic_lattice, './')
    bandstructure = get_bandstructure_input_from_exciting_structure(structure)
    bs_xml = bandstructure.to_xml()

    assert bs_xml.tag == "bandstructure", 'Root tag should be "bandstructure"'

    plot1d_xml = bs_xml.find("plot1d")
    assert plot1d_xml is not None, 'Missing "plot1d" subtree'

    path_xml = plot1d_xml.find("path")
    assert path_xml is not None, 'Missing "path" subtree'
    assert path_xml.get("steps") == "100", 'Invalid value for "steps" attribute'

    assert len(list(path_xml)) == 8
    point1 = path_xml[0]
    assert point1.get("coord") == "0.0 0.0 0.0", 'Invalid value for "coord" attribute of point 1'
    assert point1.get("label") == "G", 'Invalid value for "label" attribute of point 1'

    point2 = path_xml[1]
    assert point2.get("coord") == "0.0 0.5 0.0", 'Invalid value for "coord" attribute of point 2'
    assert point2.get("label") == "X", 'Invalid value for "label" attribute of point 2'

    point3 = path_xml[2]
    assert point3.get("coord") == "0.5 0.5 0.0", 'Invalid value for "coord" attribute of point 3'
    assert point3.get("label") == "M", 'Invalid value for "label" attribute of point 3'

    point4 = path_xml[3]
    assert point4.get("coord") == "0.0 0.0 0.0", 'Invalid value for "coord" attribute of point 4'
    assert point4.get("label") == "G", 'Invalid value for "label" attribute of point 4'

    point5 = path_xml[4]
    assert point5.get("coord") == "0.5 0.5 0.5", 'Invalid value for "coord" attribute of point 5'
    assert point5.get("label") == "R", 'Invalid value for "label" attribute of point 5'

    point6 = path_xml[5]
    assert point6.get("coord") == "0.0 0.5 0.0", 'Invalid value for "coord" attribute of point 6'
    assert point6.get("label") == "X", 'Invalid value for "label" attribute of point 6'
    assert point6.get("breakafter") == "true", 'Invalid value for "breakafter" attribute of point 6'

    point7 = path_xml[6]
    assert point7.get("coord") == "0.5 0.5 0.0", 'Invalid value for "coord" attribute of point 7'
    assert point7.get("label") == "M", 'Invalid value for "label" attribute of point 7'

    point8 = path_xml[7]
    assert point8.get("coord") == "0.5 0.5 0.5", 'Invalid value for "coord" attribute of point 8'
    assert point8.get("label") == "R", 'Invalid value for "label" attribute of point 8'



""" Tests for utils. """
from excitingtools.utils.utils import can_be_float, convert_to_literal, get_new_line_indices, flatten_list


def test_can_be_float():
    """
    Test can_be_float function on different input types.
    """
    assert can_be_float('1.0'), (
        "Expect string of a literal '1.0' can convert to float")
    assert can_be_float('1'), (
        "Expect string of a literal '1' can convert to float")
    assert can_be_float(True), (
        "Expect True can be converted to a float (would be 1.0)")
    assert can_be_float('a') is False, (
        "Expect string of the letter 'a' cannot be converted to a float")


def test_convert_to_literal():
    """
    Test convert_to_literal turns string reps of numeric data into numeric data.
    """
    assert convert_to_literal('1.1') == 1.1, "string of literal '1.1' converts to float"
    assert convert_to_literal('1.0') == 1.0, "string of literal '1.0' converts to float"
    assert convert_to_literal('1') == 1, "string of literal '1' converts to int"


def test_get_new_line_indices():
    """
    Test getting new line indices function.
    """
    test_string = 'Test Here\n 2nd Line'
    expected_line_indices = 2
    expected_line_index_list = [0, 10]
    assert len(get_new_line_indices(test_string)) == expected_line_indices
    assert get_new_line_indices(test_string)[0] == expected_line_index_list[0]
    assert get_new_line_indices(test_string)[1] == expected_line_index_list[1]


def test_flatten_list():
    input_list = [[1, 2, 3], 4, 5, [6, [7, 8]], {'9': 9, '10': 10}]
    ref_list = [1, 2, 3, 4, 5, 6, 7, 8, {'9': 9, '10': 10}]
    assert list(flatten_list(input_list)) == ref_list



""" Eigenvalue parser, processing dict values and returning to an object.
"""

import numpy as np
from excitingtools.exciting_dict_parsers.groundstate_parser import parse_eigval
from excitingtools.dataclasses.eigenvalues import EigenValues
from excitingtools.dataclasses.data_structs import NumberOfStates


def parse_eigenvalues(root) -> EigenValues:
    """ High-level parser for eigenvalues. Calls dictionary parser to parse information from the "eigval.xml" file.

    :param root: XML file name, XML string or ElementTree.Element as input.
    :return: EigenValues object.
    """

    eigval_dict = parse_eigval(root)
    data = eigval_dict['kpt']

    k_points = []
    for point in data.values():
        k_points.append([float(x) for x in point['vkl'].split()])

    k_indices = list(map(int, data.keys()))

    n_k = len(k_points)
    n_e = len(data['1']['state'])
    eigenvalues = np.empty((n_k, n_e))
    occupations = np.empty((n_k, n_e))

    for ik in range(n_k):
        for ie in range(n_e):
            eigenvalues[ik, ie] = data[str(ik + 1)]['state'][str(ie + 1)]['eigenvalue']
            occupations[ik, ie] = data[str(ik + 1)]['state'][str(ie + 1)]['occupancy']

    return EigenValues(state_range=NumberOfStates(1, occupations.shape[1]),
                       k_points=k_points,
                       k_indices=k_indices,
                       all_eigenvalues=eigenvalues,
                       occupations=occupations)



"""Test parsers that comprise the GW_INFO.OUT parser.

One notes that GW_INFO.OUT is plain text, and so any change to the formatting
or text which is used in the parser pattern-matching, will break the parser.
A much better format for this file would be YAML.
"""

import numpy as np

from excitingtools.exciting_dict_parsers.gw_info_parser import parse_gw_info, parse_frequency_grid, \
    parse_correlation_self_energy_params, extract_kpoints, parse_ks_eigenstates, \
    parse_n_q_point_cycles, parse_band_structure_info, parse_mixed_product_params, \
    parse_bare_coulomb_potential_params, parse_gw_timings

# Text files are large, it's easier to store externally.
# Note, these fixtures are used, even if greyed out by the IDE
from . mock_gw_info_out import zro2_gw_info_out_mock, si_2_gw_info_out_mock


def test_parse_correlation_self_energy_params(zro2_gw_info_out_mock):
    """ Test `Correlation self-energy parameters` block of GW_INFO.OUT.
    """
    reference = {'Solution of the QP equation': 0,
                 'Energy alignment': 0,
                 'Analytic continuation method': "PADE - Thiele's reciprocal difference method",
                 'Analytic continuation method citation': "H. J. Vidberg and J. W. Serence, J. Low Temp. Phys. 29, 179 (1977)",
                 'Scheme to treat singularities': 'Auxiliary function method "mpb"',
                 'Scheme to treat singularities citation': 'S. Massidda, M. Posternak, and A. Baldereschi, PRB 48, 5058 (1993)'
                 }

    output = parse_correlation_self_energy_params(zro2_gw_info_out_mock.string)
    assert reference == output, "parse_correlation_self_energy_params dictionary not consistent with reference"


def test_parse_mixed_product_params(zro2_gw_info_out_mock):
    """ Test `Mixed product basis parameters` block of GW_INFO.OUT.
    """
    ref = {
        'MT Angular momentum cutoff': 4,
        'MT Linear dependence tolerance factor': 0.001,
        'Plane wave cutoff (in units of Gkmax)': 1.0
    }

    output = parse_mixed_product_params(zro2_gw_info_out_mock.string)

    assert output == ref, "Expect parsed mixed product basis parameters to equal the reference"


def test_parse_bare_coulomb_potential_params(zro2_gw_info_out_mock):
    """ Test `Bare Coulomb potential parameters` block of GW_INFO.OUT.
    """
    ref = {
        'Plane wave cutoff (in units of Gkmax*gmb)': 2.0,
        'Error tolerance for structure constants': 1e-16,
        'MB tolerance factor': 0.1
    }

    output = parse_bare_coulomb_potential_params(zro2_gw_info_out_mock.string)

    assert output == ref, "Expect parsed bare Coulomb parameters to match the reference"


def test_parse_frequency_grid(zro2_gw_info_out_mock):
    """ Test parsing `frequency grid` block of GW_INFO.OUT.
    """
    n_points = 32
    f_grid = parse_frequency_grid(zro2_gw_info_out_mock.string, n_points)

    ref_frequencies = np.array([5.2995325042E-03, 2.7712488463E-02, 6.7184398806E-02, 0.1222977958, 0.1910618778,
                                0.2709916112, 0.3591982246, 0.4524937451, 0.5475062549, 0.6408017754,
                                0.7290083888, 0.8089381222, 0.8777022042, 0.9328156012, 0.9722875115,
                                0.9947004675, 1.005327767, 1.028502360, 1.072023237, 1.139338599,
                                1.236188495, 1.371726328, 1.560544990, 1.826463152, 2.209975300,
                                2.783978125, 3.690151129, 5.233906478, 8.176762249, 14.88440795,
                                36.08481430, 188.6958895])

    ref_weights = np.array([1.3576229706E-02, 3.1126761969E-02, 4.7579255841E-02, 6.2314485628E-02, 7.4797994408E-02,
                            8.4578259698E-02, 9.1301707522E-02, 9.4725305228E-02, 9.4725305228E-02, 9.1301707522E-02,
                            8.4578259698E-02, 7.4797994408E-02, 6.2314485628E-02, 4.7579255841E-02, 3.1126761969E-02,
                            1.3576229706E-02, 1.3721277051E-02, 3.2926421206E-02, 5.4679689940E-02, 8.0889962951E-02,
                            0.1143034524, 0.1591452545, 0.2223471090, 0.3160005534, 0.4626375217,
                            0.7076370069, 1.151720377, 2.048999581, 4.166311667, 10.54097479,
                            40.53058703, 483.3971183])

    assert len(ref_frequencies) == 32, "Require 32 reference frequency points"
    assert len(ref_weights) == 32, "Require 32 reference weights"

    assert np.allclose(f_grid[0, :],
                       ref_frequencies), "Frequency points parsed from gw_info_out disagree with reference"
    assert np.allclose(f_grid[1, :], ref_weights), "Weights parsed from gw_info_out disagree with reference"


def test_parse_ks_eigenstates(zro2_gw_info_out_mock):
    """ Test parsing ` Kohn-Sham eigenstates summary` block from GW_INFO.OUT.
    """
    ref = {
        'Maximum number of LAPW states': 847,
        'Minimal number of LAPW states': 838,
        'Number of states used in GW - total KS': 838,
        'Number of states used in GW - occupied': 21,
        'Number of states used in GW - unoccupied': 2000,
        'Number of states used in GW - dielectric function': 838,
        'Number of states used in GW - self energy': 838,
        'Energy of the highest unoccupied state': 1030.791933,
        'Number of valence electrons': 42,
        'Number of valence electrons treated in GW': 42
    }

    output = parse_ks_eigenstates(zro2_gw_info_out_mock.string)

    assert output == ref, "Parsed KS eigenstates settings not consistent with reference"


def test_parse_n_q_point_cycles(zro2_gw_info_out_mock):
    max_q = parse_n_q_point_cycles(zro2_gw_info_out_mock.string)
    assert max_q == 2, "Two q cycles expected from the reference data"


def test_extract_kpoint(zro2_gw_info_out_mock):
    ref = {'VBM': {'k_point': [0.0, 0.5, 0.5], 'ik': 3},
           'CBm': {'k_point': [0.0, 0.0, 0.0], 'ik': 1}
           }

    output = extract_kpoints(zro2_gw_info_out_mock.string)

    assert output == ref, "Expect extracted VBM and CBm k-points to match reference"


def test_parse_band_structure_info(zro2_gw_info_out_mock):
    """ Test parsing the `Kohn-Sham band structure` block of GW_INFO.OUT
    """
    ks_ref = {
        'Fermi energy': 0.0,
        'Energy range': [-14.6863, 1030.7919],
        'Band index of VBM': 21,
        'Band index of CBm': 22,
        'Indirect BandGap (eV)': 3.3206,
        'Direct Bandgap at k(VBM) (eV)': 3.7482,
        'Direct Bandgap at k(CBm) (eV)': 3.8653,
        'VBM': {'k_point': [0.0, 0.5, 0.5], 'ik': 3},
        'CBm': {'k_point': [0.0, 0.0, 0.0], 'ik': 1}
    }
    ks_output = parse_band_structure_info(zro2_gw_info_out_mock.string, 'ks')
    assert ks_output == ks_ref, "Expect parsed KS band structure info to match the reference"


def test_parse_g0w0_band_structure_info(si_2_gw_info_out_mock, zro2_gw_info_out_mock):
    """ Test parsing the `G0W0 band structure` block of GW_INFO.OUT
    """
    # Direct gap
    g0w0_ref = {
        'Fermi energy': 0.0176,
        'Energy range': [-0.4799, 0.5045],
        'Band index of VBM': 4,
        'Band index of CBm': 5,
        'Direct BandGap (eV)': 3.2457,
        'VBM': {'k_point': [0.0, 0.0, 0.0], 'ik': 1},
        'CBm': {'k_point': [0.0, 0.0, 0.0], 'ik': 1}
    }
    gw_output = parse_band_structure_info(si_2_gw_info_out_mock.string, 'gw')
    assert gw_output == g0w0_ref, \
        "Expect parsed G0W0 band structure info to match the reference for direct gap "


    # Indirect gap
    g0w0_ref = {
        'Fermi energy': -0.0054,
        'Energy range': [-16.2632, 1031.409],
        'Band index of VBM': 21,
        'Band index of CBm': 22,
        'Indirect BandGap (eV)': 5.392,
        'Direct Bandgap at k(VBM) (eV)': 5.5472,
        'Direct Bandgap at k(CBm) (eV)': 5.9646,
        'VBM': {'k_point': [0.0, 0.5, 0.5], 'ik': 3},
        'CBm': {'k_point': [0.0, 0.0, 0.0], 'ik': 1}
    }
    gw_output = parse_band_structure_info(zro2_gw_info_out_mock.string, 'gw')
    assert gw_output == g0w0_ref, \
        "Expect parsed G0W0 band structure info to match the reference for indirect gap"


def test_parse_gw_info(zro2_gw_info_out_mock):
    """ Test parsing of the whole GW_INFO.OUT
    """

    # Reference, without frequencies_weights
    ref = {'correlation_self_energy_parameters': {'Solution of the QP equation': 0,
                                                  'Energy alignment': 0,
                                                  'Analytic continuation method': "PADE - Thiele's reciprocal difference method",
                                                  'Analytic continuation method citation':
                                                      'H. J. Vidberg and J. W. Serence, J. Low Temp. Phys. 29, 179 (1977)',
                                                  'Scheme to treat singularities': 'Auxiliary function method "mpb"',
                                                  'Scheme to treat singularities citation':
                                                      'S. Massidda, M. Posternak, and A. Baldereschi, PRB 48, 5058 (1993)'},
           'mixed_product_basis_parameters': {'MT Angular momentum cutoff': 4,
                                              'MT Linear dependence tolerance factor': 0.001,
                                              'Plane wave cutoff (in units of Gkmax)': 1.0},
           'bare_coulomb_potential_parameters': {'Plane wave cutoff (in units of Gkmax*gmb)': 2.0,
                                                 'Error tolerance for structure constants': 1e-16,
                                                 'MB tolerance factor': 0.1},
           'screened_coulomb_potential': 'Full-frequency Random-Phase Approximation',
           'core_electrons_treatment': 'all - Core states are included in all calculations',
           'qp_interval': [1, 2000],
           'n_empty': 2000,
           'q_grid': [2, 2, 2],
           'mixed_product_wf_info': {'Maximal number of MT wavefunctions per atom': 1069,
                                     'Total number of MT wavefunctions': 2733,
                                     'Maximal number of PW wavefunctions': 468,
                                     'Total number of mixed-product wavefunctions': 3201},
           'frequency_grid': {'Type: < fgrid >': 'gauleg2',
                              'Frequency axis: < fconv >': 'imfreq',
                              'Number of frequencies: < nomeg >': 32,
                              'Cutoff frequency: < freqmax >': 1.0},
           'ks_eigenstates_summary': {'Maximum number of LAPW states': 847,
                                      'Minimal number of LAPW states': 838,
                                      'Number of states used in GW - total KS': 838,
                                      'Number of states used in GW - occupied': 21,
                                      'Number of states used in GW - unoccupied': 2000,
                                      'Number of states used in GW - dielectric function': 838,
                                      'Number of states used in GW - self energy': 838,
                                      'Energy of the highest unoccupied state': 1030.791933,
                                      'Number of valence electrons': 42,
                                      'Number of valence electrons treated in GW': 42},
           'ks_band_structure_summary': {'Fermi energy': 0.0,
                                         'Energy range': [-14.6863, 1030.7919],
                                         'Band index of VBM': 21, 'Band index of CBm': 22,
                                         'Indirect BandGap (eV)': 3.3206,
                                         'Direct Bandgap at k(VBM) (eV)': 3.7482,
                                         'Direct Bandgap at k(CBm) (eV)': 3.8653,
                                         'VBM': {'k_point': [0.0, 0.5, 0.5], 'ik': 3},
                                         'CBm': {'k_point': [0.0, 0.0, 0.0], 'ik': 1}},
           'n_q_cycles': 2,
           'g0w0_band_structure_summary': {'Fermi energy': -0.0054,
                                           'Energy range': [-16.2632, 1031.409],
                                           'Band index of VBM': 21,
                                           'Band index of CBm': 22,
                                           'Indirect BandGap (eV)': 5.392,
                                           'Direct Bandgap at k(VBM) (eV)': 5.5472,
                                           'Direct Bandgap at k(CBm) (eV)': 5.9646,
                                           'VBM': {'k_point': [0.0, 0.5, 0.5], 'ik': 3},
                                           'CBm': {'k_point': [0.0, 0.0, 0.0], 'ik': 1}}
           }

    output = parse_gw_info(zro2_gw_info_out_mock.file)

    # frequencies and weights tested in separate unit test
    f_w = output['frequency_grid'].pop('frequencies_weights')

    assert output == ref, "Output from parse_gw_info does not agree with reference dictionary"


def test_parse_gw_timings(zro2_gw_info_out_mock):
    """ Test parsing `GW timing info` block in GW_INFO.OUT
    """

    ref = {
        'Initialization': {
            'Initialization': 15.46,
            'init_scf': 8.38,
            'init_kpt': 0.04,
            'init_eval': 0.02,
            'init_freq': 0.0,
            'init_mb': 6.76
        },
        'Subroutines': {
            'Subroutines': None,
            'calcpmat': 5.12,
            'calcbarcmb': 5.65,
            'BZ integration weights': 18.35
        },
        'Dielectric function': {
            'Dielectric function': 422.09,
            'head': 0.3,
            'wings': 70.14,
            'body (not timed)': 0.0,
            'inversion': 1.72
        },
        'WF products expansion': {
            'WF products expansion': 2525.59,
            'diagsgi': 0.24,
            'calcmpwipw': 0.04,
            'calcmicm': 2.63,
            'calcminc': 2.1,
            'calcminm': 2520.58
        },
        'Self-energy': {'Self-energy': 7069.46, 'calcselfx': 43.33, 'calcselfc': 7026.14},
        'calcvxcnn': {'calcvxcnn': 27.52},
        'input/output': {'input/output': 0.0},
        'Total': {'Total': 7555.78}
    }

    assert parse_gw_timings(zro2_gw_info_out_mock.string) == ref, "Parsed timings do not agree with reference dictionary"



""" Eigenvalue class.
"""
import warnings
from itertools import permutations
from typing import List, Union, Optional

import numpy as np
from excitingtools.dataclasses.data_structs import PointIndex, BandIndices, NumberOfStates


class EigenValues:
    point_type = Union[np.ndarray, List[float]]
    index_type = Union[np.ndarray, List[int]]
    # If a k-index is not matched
    NO_MATCH = -1

    def __init__(self,
                 state_range: NumberOfStates,
                 k_points: point_type,
                 k_indices: index_type,
                 all_eigenvalues: np.ndarray,
                 weights=None,
                 occupations: Optional[np.ndarray] = None):
        self.state_range = state_range
        self.k_points = k_points
        self.k_indices = k_indices
        self.all_eigenvalues = all_eigenvalues
        self.weights = weights
        self.occupations = occupations
        if all_eigenvalues.shape != (len(self.k_points), self.state_range.n_states):
            raise ValueError('Shape of all_eigenvalues does not match (n_k, n_states)')

    def get_array_index(self, i_state: int):
        """ Given the state index, get the corresponding index in the eigenvalue array.

        :param i_state: State index using fortran indexing
        :return array index, using zero-indexing
        """
        assert i_state > 0, "state indexing starts at 1"
        return i_state - self.state_range.first_state

    def get_k_point(self, ik: int):
        """ Get the k-point associated with ik index.

        :param ik: k-point index.
        :return k-point in fractional coordinates.
        """
        assert ik > 0, "ik indexing starts at 1"
        return self.k_points[ik - 1]

    def get_index(self, k_point, verbose=False) -> int:
        """ Find the corresponding index of a k-point.

        If no k-point is found, NO_MATCH is returned.

        :param k_point: k-point in fractional coordinates.
        :param verbose: Print warning, if no k-point found.
        :return ik: Corresponding index w.r.t. exciting.
        """
        diff = np.empty(shape=len(self.k_points))
        k_point = np.asarray(k_point)
        for i, point in enumerate(self.k_points):
            diff[i] = np.linalg.norm(np.asarray(point) - k_point)
        indices = np.argwhere(diff < 1.e-8)

        if len(indices) == 0:
            if verbose:
                warnings.warn(f'{k_point} not present in list of k-points')
            return self.NO_MATCH

        indices = indices[0]

        if len(indices) > 1:
            raise ValueError(f'Found degenerate k-points at {indices}')

        ik = indices[0] + 1
        return ik

    def get_k_points(self) -> List[PointIndex]:
        """K-points and their indices

        Note. Fortran indexing for ik always input and output.
        """
        k_points_and_indices = []
        for ik in self.k_indices:
            k_point = self.get_k_point(ik)
            k_points_and_indices.append(PointIndex(k_point, ik))
        return k_points_and_indices

    def get_eigenvalues(self, ik: Optional[int] = None, k_point=None) -> np.ndarray:
        """Return eigenvalues for column="data" w.r.t. k-point.

        :param ik: k-point index.
        :param k_point: k-point, used to find the k-index if ik is not passed.
        :return eigenvalues at ik k-point.
        """
        if ik is None:
            if k_point is None:
                raise ValueError('Must provide either k-index or k-point')
            ik = self.get_index(k_point)

            if ik == self.NO_MATCH:
                return np.empty(shape=0)

        return self.all_eigenvalues[ik - 1, :]

    def band_gap(self, band_indices: BandIndices, k_points=None, k_indices=None) -> Union[float, ValueError]:
        """ Get a band gap for two k-points in the valence band top
        and conduction band bottom, respectively.

        TODO(Alex) Consider adding objects to hold k_points and k_indices.
        However, one can retain support for tuples/lists.

        :param band_indices: Band indices
        :param k_points: k-points for the valence and conduction bands, in that order.
        :param k_indices: k-indices for the valence and conduction bands, in that order.
        :return Band gap. A ValueError is returned if one or both k-points are not present.
        """
        if k_indices is None:
            if k_points is None:
                raise ValueError('Must pass either k_points or k_indices to band_gap method')
            k_indices = (self.get_index(k_points[0]), self.get_index(k_points[1]))

        if self.NO_MATCH in k_indices:
            indices = np.argwhere(k_indices == self.NO_MATCH)[0]
            err_msg = f"".join(f'Requested k-point {k_points[i]} not present. \n' for i in indices)
            return ValueError(err_msg)

        i_vbm = self.get_array_index(band_indices.VBM)
        i_cbm = self.get_array_index(band_indices.CBm)

        ik_vbm = k_indices[0] - 1
        ik_cbm = k_indices[1] - 1

        return self.all_eigenvalues[ik_cbm, i_cbm] - self.all_eigenvalues[ik_vbm, i_vbm]

    def get_transition_energy(self, valence_k_point: point_type, conduction_k_point: point_type) -> float:
        """Determine transition energy between two k-points in the valence band top and conduction band bottom,
        respectively.

        This function accounts for all different permutations in which the k-points can be found within the exciting
        eigenvalue output files and determines band indices using occupation values.

        :param valence_k_point: k-point for valence band in fractional coordinates.
        :param conduction_k_point: k-point for conduction band in fractional coordinates.
        :return: Transition energy in Hartree.
        """
        indices_valence = [self.get_index(k_point) for k_point in permutations(valence_k_point)]
        try:
            ik = next(i for i in indices_valence if i != self.NO_MATCH)
        except StopIteration:
            raise ValueError(f'Requested valence k-point {valence_k_point} not present.')

        indices_conduction = [self.get_index(k_point) for k_point in permutations(conduction_k_point)]
        try:
            jk = next(i for i in indices_conduction if i != self.NO_MATCH)
        except StopIteration:
            raise ValueError(f'Requested conduction k-point {conduction_k_point} not present.')

        iVBM = np.where(self.occupations[ik - 1] == 0)[0][0] + self.state_range.first_state - 1
        jCBm = np.where(self.occupations[jk - 1] == 0)[0][0] + self.state_range.first_state

        return self.band_gap(BandIndices(iVBM, jCBm), k_indices=[ik, jk])




""" Function for finding the corresponding index of a k-point.
"""
import numpy as np
import warnings


def get_k_point_index(k_point: np.ndarray, k_points_to_search: np.ndarray, verbose=False) -> int:
    """ Find the corresponding index of a k-point.

    If no k-point is found, NO_MATCH is returned.

    :param k_point: k-point in fractional coordinates.
    :param k_points_to_search: Array of k-points in which k_point is searched for
    :param verbose: Print warning, if no k-point found.
    :return ik: Corresponding index w.r.t. exciting.
    """
    NO_MATCH = np.NaN

    diff = np.empty(shape=len(k_points_to_search))
    k_point = np.asarray(k_point)
    for i, point in enumerate(k_points_to_search):
        diff[i] = np.linalg.norm(np.asarray(point) - k_point)
    indices = np.argwhere(diff < 1.e-8)

    if len(indices) == 0:
        if verbose:
            warnings.warn(f'{k_point} not present in list of k-points')
        return NO_MATCH

    if len(indices) > 1:
        raise ValueError(f'Found degenerate k-points at {indices}'.replace('\n', ','))

    ik = indices[0]
    return ik



""" Test for dynamic class functions. """

from excitingtools.input.dynamic_class import class_constructor_string, class_name_uppercase, get_all_valid_subtrees, \
    give_class_dictionary


def test_get_all_valid_subtrees():
    ref_subtrees = {'crystal', 'shell', 'dfthalfparam', 'structure', 'symmetries',
                    'basevect', 'species', 'atom', 'LDAplusU'}
    assert set(get_all_valid_subtrees(["structure"])) == ref_subtrees


def test_class_dir():
    ref_dir = {'attributes': {'__doc__': 'Class for exciting structure input.',
                              '__module__': 'excitingtools.input.input_classes',
                              'name': 'structure'},
               'bases': '(ExcitingXMLInput, )'}
    assert give_class_dictionary("structure") == ref_dir


def test_class_constructor_string():
    test_input = {'Spin': {"bases": '(ExcitingXMLInput, )', "attributes": {
        "__doc__": "Class for exciting spin input.", 'name': "spin"}}}
    ref_string = "ExcitingSpinInput = type('ExcitingSpinInput', (ExcitingXMLInput, ), " \
                 "{'__doc__': 'Class for exciting spin input.', 'name': 'spin'}) \n"
    assert class_constructor_string(test_input) == ref_string


def test_class_name_uppercase():
    assert class_name_uppercase("groundstate") == "GroundState"
    assert class_name_uppercase("xs") == "XS"
    assert class_name_uppercase("structure") == "Structure"
    assert class_name_uppercase("LDAplusU") == "LDAplusU"
    assert class_name_uppercase("HartreeFock") == "HartreeFock"
    assert class_name_uppercase("EFG") == "EFG"
    assert class_name_uppercase("etCoeffComponents") == "EtCoeffComponents"



""" Parser for STATE.OUT binary file.
"""
import numpy as np
import struct
import sys


def get_byteorder_format_char(byteorder) -> str:
    """ Converts the byteorder string to formatting symbol for struct.unpack function

    :param byteorder: endianness of the byteorder ("little" or "big")
    :return: the character representing the endianness ('<' or '>')
    """
    chars = {
        "little": '<',
        "big": '>',
    }
    return chars[byteorder]


def read_array(file, shape, byteorder) -> np.ndarray:
    """ Read in a sequence of double values and reshape them into the wanted shape.

    :param file: binary file object, which contains the number sequence
    :param shape: the desired shape of the number sequence (with column-major order)
    :param byteorder: the endianness of the bytes representation
    :return: a numpy array of double values with the desired shape
    """
    # get the correct format character for the byteorder
    byteorder = get_byteorder_format_char(byteorder)
    # get the number of bytes for the array
    num_bytes = np.prod(shape) * 8
    # get all the double numbers
    values = struct.unpack(f"{byteorder}{np.prod(shape)}d", file.read(num_bytes))
    # return the values as correctly shaped numpy array
    return np.reshape(values, shape, order='F')


def read_complex_array(file, shape, byteorder) -> np.ndarray:
    """ Read in a sequence of complex values and reshape them into the wanted shape.

    :param file: binary file object, which contains the number sequence
    :param shape: the desired shape of the number sequence (with column-major order)
    :param byteorder: the endianness of the bytes representation
    :return: a numpy array of complex values with the desired shape
    """
    byteorder = get_byteorder_format_char(byteorder)
    array = np.reshape(
        [complex(*struct.unpack(f"{byteorder}2d", file.read(16))) for _ in range(np.prod(shape))], shape, order='F'
    )
    return array


def read_int(file, byteorder, num_bytes=4, signed=False) -> int:
    """ Read in a single integer.

    :param file: binary file object, which contains the number
    :param byteorder: the endianness of the bytes representation
    :param num_bytes: the number of bytes, which constitute the integer
    :param signed: weather the bytes should be interpreted as a signed or unsigned integer
    :return: the read in integer
    """
    return int.from_bytes(file.read(num_bytes), byteorder, signed=signed)


def read_integers(file, integers, dest, byteorder):
    """ Read in a sequence of named integers and store them in a dictionary.

    :param file: binary file object, which contains the integers
    :param integers: the keys of the integers for the dictionary
    :param dest: the dictionary, which stores the integers and gets modified
    :param byteorder: the endianness of the bytes representation
    """
    for integer in integers:
        num_bytes = read_int(file, byteorder)
        dest[integer] = read_int(file, byteorder, num_bytes)
        assert num_bytes == read_int(file, byteorder)


def parse_state_out(path, byteorder=sys.byteorder) -> dict:
    """ Parser for: STATE.OUT

    STATE.OUT is a binary file. For every 'Write' (in Fortran) there are 4 leading bytes, which are an integer
    equal to the number of bytes of the Fortran objects written. After this the leading 4 bytes are
    repeated.
    This file contains information about the version of exciting, which was used and all the information about
    the density and potentials.
    The functions for the muffin-tin region are stored as an expansion of real spherical harmonics.

    :param path: the path to binary file
    :param byteorder: the endianness of the file (defaults to the endianness of the system)
    :return: a dictionary containing all the state information
    """
    state = {}
    radial_meshes = []
    number_of_atoms = []
    with open(path, "rb") as file:
        # first read the version tuple (3 * 4 bytes) and the versionhash (40 bytes)
        assert read_int(file, byteorder) == 52
        state["version"] = tuple(read_int(file, byteorder) for _ in range(3))
        state["versionhash"] = file.read(40).decode("utf-8")
        assert read_int(file, byteorder) == 52
        # next read whether the calculation was spin polarized or not (4 bytes)
        assert read_int(file, byteorder) == 4
        state["spinpol"] = read_int(file, byteorder) == 1
        assert read_int(file, byteorder) == 4
        # next we read different integers
        integers = ("nspecies", "lmmaxvr", "nrmtmax")
        read_integers(file, integers, state, byteorder)
        # next for every species there are the number of atoms for this species, the number of points
        # of the radial mesh and finally the radial mesh
        for _ in range(state["nspecies"]):
            num_bytes = read_int(file, byteorder)
            number_of_atoms.append(read_int(file, byteorder, num_bytes))
            assert num_bytes == read_int(file, byteorder)
            num_bytes = read_int(file, byteorder)
            nrmt = read_int(file, byteorder, num_bytes)
            assert num_bytes == read_int(file, byteorder)
            num_bytes = read_int(file, byteorder)
            assert num_bytes // 8 == nrmt
            radial_meshes.append(read_array(file, nrmt, byteorder))
            assert num_bytes == read_int(file, byteorder)
        # save the number of atoms per species
        state["number of atoms"] = number_of_atoms
        # save the radial meshes
        state["muffintin radial meshes"] = radial_meshes

        # next read the g vector grid size (3 * 4 bytes)
        num_bytes = read_int(file, byteorder)
        state["g vector grid"] = tuple(read_int(file, byteorder) for _ in range(3))
        assert num_bytes == read_int(file, byteorder)

        # next again read a few integers
        integers = ("ngvec", "ndmag", "nspinor", "ldapu", "lmmaxlu")
        read_integers(file, integers, state, byteorder)

        # next read a few pairs of arrays
        arrays = (
            ("muffintin density", "interstitial density"),
            ("muffintin coulomb potential", "interstitial coulomb potential"),
            ("muffintin exchange-correlation potential", "interstitial exchange-correlation potential"),
        )
        muffintin_shape = (state["lmmaxvr"], state["nrmtmax"], sum(number_of_atoms))
        interstitial_shape = np.prod(state["g vector grid"])
        for muffintin_array, interstitial_array in arrays:
            num_bytes = read_int(file, byteorder)
            assert num_bytes // 8 == (np.prod(muffintin_shape) + interstitial_shape)
            state[muffintin_array] = read_array(file, muffintin_shape, byteorder)
            state[interstitial_array] = read_array(file, interstitial_shape, byteorder)
            assert num_bytes == read_int(file, byteorder)
        # the next entry is an array triple
        num_bytes = read_int(file, byteorder)
        assert num_bytes // 8 == (np.prod(muffintin_shape) + interstitial_shape + 2 * state["ngvec"])
        state["muffintin effective potential"] = read_array(file, muffintin_shape, byteorder)
        state["interstitial effective potential"] = read_array(file, interstitial_shape, byteorder)
        state["reciprocal interstitial effective potential"] = read_complex_array(file, state["ngvec"], byteorder)
        assert num_bytes == read_int(file, byteorder)

        # the next two arrays are only present if the calculation was spin polarized
        if state["spinpol"]:
            num_bytes = read_int(file, byteorder)
            state["muffintin magnetization"] = read_array(file, muffintin_shape, byteorder)
            state["interstitial magnetization"] = read_array(file, interstitial_shape, byteorder)
            assert num_bytes == read_int(file, byteorder)

        if state["ldapu"] != 0:
            shape = (state["lmmaxlu"], state["lmmaxlu"], state["nspinor"], state["nspinor"], sum(number_of_atoms))
            num_bytes = read_int(file, byteorder)
            assert num_bytes // 8 == 2 * np.prod(shape)
            state["vmatlu"] = read_complex_array(file, shape, byteorder)
            assert num_bytes == read_int(file, byteorder)

    return state



""" Parse the schema and generate a python file which can be read by the input classes.
Should only be run if changes to the schema are made.
"""
import re
from pathlib import Path
from typing import List

import xmlschema
from xmlschema.validators import XsdAnyElement

from excitingtools.utils.utils import get_excitingtools_root


def copy_schema_files_for_parsing(schema_files: List[str]) -> List[Path]:
    """ Copies the schema files to the current directory.

    Also generates a file (input.xsd) with schema extensions, with modified include paths.
    This is a work-around because the exciting documentation fails to compile
    when a schema with the modified include is used.

    :param schema_files: the name of the schema files and tags of the xml element
    :return file_list: A list of all generated or copied files.
    """
    inputschemaextentions_name = "inputschemaextentions"
    inputschemaextentions = f"""<?xml version="1.0" encoding="UTF-8"?>
<xs:schema xmlns:ex="{inputschemaextentions_name}.xsd"
xmlns:xs="http://www.w3.org/2001/XMLSchema"
targetNamespace="{inputschemaextentions_name}.xsd"
xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
 >
 
  <xs:attribute name="importance">
    <xs:simpleType>
      <xs:restriction base="xs:string">
        <xs:enumeration value="essential"></xs:enumeration>
        <xs:enumeration value="expert"></xs:enumeration>
        <xs:enumeration value="experimental"></xs:enumeration>
        <xs:enumeration value="output"></xs:enumeration>
        <xs:enumeration value="ignore"></xs:enumeration>
      </xs:restriction>
    </xs:simpleType>
  </xs:attribute>
  
  <xs:attribute name="unit">
    <xs:simpleType>
      <xs:restriction base="xs:string">
        <xs:enumeration value=""></xs:enumeration>
        <xs:enumeration value="1"></xs:enumeration>
        <xs:enumeration value="rad"></xs:enumeration>
        <xs:enumeration value="Bohr"></xs:enumeration>
        <xs:enumeration value="Hartree"></xs:enumeration>
        <xs:enumeration value="atomic units"></xs:enumeration>
        <xs:enumeration value="lattice coordinates"></xs:enumeration>
      </xs:restriction>
    </xs:simpleType>
  </xs:attribute>
  
</xs:schema>"""

    inputschemaextentions_path = Path(f'{inputschemaextentions_name}.xsd')
    inputschemaextentions_path.write_text(inputschemaextentions)
    root = get_excitingtools_root()

    # handling of the input.xsd file
    inputschema_file = root / '../../xml/schema/input.xsd'
    content = inputschema_file.read_text().split("\n")
    content[1] = ('<xs:schema xmlns:ex="inputschemaextentions.xsd" '
                  'xmlns:xs="http://www.w3.org/2001/XMLSchema" '
                  'xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" '
                  'xsi:schemaLocation="inputschemaextentions.xsd inputschemaextentions.xsd">')
    new_inputschema_file = Path("input.xsd")
    new_inputschema_file.write_text("\n".join(content))

    file_list = [inputschemaextentions_path, new_inputschema_file]

    # handling of all other included schema files
    for schema_name in schema_files:
        schema_root = root / '../../xml/schema'
        schema_reference = schema_root / f'{schema_name}.xsd'
        content = schema_reference.read_text().split("\n")

        content[1] = f'  xmlns:ex="{inputschemaextentions_name}.xsd"'
        content[3] = f'  xsi:schemaLocation="{inputschemaextentions_name}.xsd {inputschemaextentions_name}.xsd">'
        content = "\n".join(content)

        new_schema_file = Path(f"{schema_name}.xsd")
        new_schema_file.write_text(content)
        file_list.append(new_schema_file)

    return file_list


def read_schema_to_dict(name: str) -> dict:
    """ Read schema and transform to sensible dictionary.

    Note: This could be done with an external library, such as `xmltodict` or `xmljson`
    but as this module should be run infrequently, a custom implementation reduces
    dependencies.

    :param name: name of the schema file and tag of the xml element
    :return: a dictionary with the need information about children/parents and valid attributes.
    """
    schema = xmlschema.XMLSchema(f'{name}.xsd')

    tag_info = {}
    xsd_elements = filter(lambda x: isinstance(x, xmlschema.XsdElement) and x.ref is None, schema.iter_components())

    for xsd_element in xsd_elements:
        attributes = xsd_element.attributes
        mandatory_attributes = set([k for k, v in attributes.items() if v.use == "required"])
        children = [x.name for x in xsd_element.iterchildren() if not isinstance(x, XsdAnyElement)]
        mandatory_children = set([x.name for x in xsd_element.iterchildren() if x.min_occurs > 0])
        multiple_childs = set([x.name for x in xsd_element.iterchildren() if x.max_occurs is None or x.max_occurs > 1])

        tag_info[xsd_element.name] = {"attribs": filter(lambda x: x is not None, attributes), "children": children,
                                      "mandatory_attribs": mandatory_attributes | mandatory_children,
                                      "multiple_children": multiple_childs}

        # special handling for the plan
        if xsd_element.name == "doonly":
            tag_info["doonly"]["plan"] = attributes["task"].type.validators[0].enumeration

    # exclude special structure attributes (already explicitly specified in the __init__ of ExcitingStructure class)
    if name == "structure":
        tag_info["structure"]["mandatory_attribs"].remove("crystal")
        tag_info["crystal"]["mandatory_attribs"].remove("basevect")
        tag_info["species"]["mandatory_attribs"].remove("atom")

    return tag_info


def write_schema_info(super_tag: str, schema_dict: dict) -> str:
    """ Converts dict representation of the schema to string.

    :param super_tag: name of the top-level element
    :param schema_dict: contains all the information read from the schema
    :return info_string: string of python-formatted code
    """
    info_string = f"\n# {super_tag} information \n"
    for tag in schema_dict:
        valid_attributes = sorted(schema_dict[tag]["attribs"])
        valid_subtrees = schema_dict[tag]['children']
        mandatory_attributes = sorted(schema_dict[tag]['mandatory_attribs'])
        multiple_childs = sorted(schema_dict[tag]['multiple_children'])

        if not (valid_attributes or valid_subtrees or mandatory_attributes):
            continue

        if valid_attributes:
            info_string += list_string_line_limit(f"{tag}_valid_attributes", valid_attributes) + " \n"
        if valid_subtrees:
            info_string += list_string_line_limit(f"{tag}_valid_subtrees", valid_subtrees) + " \n"
        if mandatory_attributes:
            info_string += list_string_line_limit(f"{tag}_mandatory_attributes", mandatory_attributes) + " \n"
        if multiple_childs:
            info_string += list_string_line_limit(f"{tag}_multiple_children", multiple_childs) + " \n"
        info_string += "\n"
    return info_string


def list_string_line_limit(name: str, content: list, max_length: int = 120) -> str:
    """ Given a list of unique items, produces a python formatted string containing the definition of the list
    given by the name:
        name = ['entry1', 'entry2', ...]
    Inserts a line break every time the string gets longer then the limit.

    :param name: the name of the set in the final string representing the definition
    :param content: the list to write to string as a set (list because of consistent ordering)
    :param max_length: the maximum line length
    :return: the formatted string with fixed line length
    """
    formatted_string = ''
    current_line_string = name + ' = ['
    start_len = len(current_line_string)
    entry_break_string = "', "

    for entry in content:
        if len(str(entry) + current_line_string + entry_break_string) > max_length:
            formatted_string += current_line_string + '\n'
            current_line_string = start_len * ' '
        current_line_string += f"'{entry}{entry_break_string}"

    formatted_string += current_line_string[:-2] + "]"
    return formatted_string


def get_all_include_files() -> list:
    """ Gets a list of all included files in the input.xsd file.
    """
    input_schema_file = (get_excitingtools_root() / '../../xml/schema/input.xsd').resolve()
    if not input_schema_file.exists():
        raise ValueError("Couldn't find exciting schema. Most likely you are using excitingtools outside of exciting."
                         "To fix this, try installing excitingtools from source in editable (-e) mode.")
    return re.findall(r'<xs:include id=".*" schemaLocation="(.*)\.xsd"/>', input_schema_file.read_text())


def main():
    """ Main function to read the schema and write it to python readable file.
    """
    filename = Path(__file__).parent / "valid_attributes.py"
    info = '""" Automatically generated file with the valid attributes from the schema. \n' \
           'Do not manually change. Instead, run "utils/schema_parsing.py" to regenerate. """ \n'

    schemas = get_all_include_files()
    tmp_files = copy_schema_files_for_parsing(schemas)

    input_schema_dict = {"input": read_schema_to_dict("input")["input"]}
    info += write_schema_info("input", input_schema_dict)

    for schema in schemas:
        schema_dict = read_schema_to_dict(schema)
        info += write_schema_info(schema, schema_dict)

    # Handle special case for 'xs' to handle the valid plan entries
    xs_schema_dict = read_schema_to_dict("xs")
    info += "\n# valid entries for the xs subtree 'plan'\n"
    info += list_string_line_limit("valid_plan_entries", sorted(xs_schema_dict['doonly']['plan'])) + " \n"

    with open(filename, "w") as fid:
        fid.write(info)

    for file in tmp_files:
        file.unlink()


if __name__ == "__main__":
    main()



""" Utilities to interact with the pymatgen library. """

from pymatgen.core import Structure

from excitingtools import ExcitingStructure
from excitingtools.constants.units import angstrom_to_bohr, bohr_to_angstrom


def exciting_structure_to_pymatgen(structure: ExcitingStructure) -> Structure:
    """ Function to extract the physical structure from an exciting structure object
    and transforms it into a pymatgen.core.structure.Structure object.

    :param structure: input exciting structure object
    :returns: pymatgen Structure object
    """
    lattice = structure.get_lattice(convert_to_angstrom=True)
    cartesian = structure.structure_attributes.get("cartesian", False)
    positions = structure.positions * bohr_to_angstrom if cartesian else structure.positions
    return Structure(lattice=lattice, species=structure.species, coords=positions, coords_are_cartesian=cartesian)


def pymatgen_to_exciting_structure(structure: Structure) -> ExcitingStructure:
    """ Initialise lattice, species and positions from a pymatgen Structure Object.
    Note: pymatgen works in Angstrom, whereas exciting expects atomic units

    :param structure: pymatgen Structure object.
    :return exciting structure object
    """
    lattice = structure.lattice.matrix * angstrom_to_bohr
    species = [x.symbol.capitalize() for x in structure.species]
    positions = structure.frac_coords
    atoms = [{"species": atom, "position": positions[i]} for i, atom in enumerate(species)]
    return ExcitingStructure(atoms, lattice)



""" Parsers for input.xml.

TODO(Fabian): Issues 117 & 121:
As more sub-elements are implemented in the input files, also add parsers here
"""
import copy
from typing import Tuple
from xml.etree import ElementTree

from excitingtools.parser_utils.parser_decorators import xml_root
from excitingtools.parser_utils.parser_utils import find_element, convert_string_dict
from excitingtools.utils import valid_attributes as all_valid_attributes
from excitingtools.utils.valid_attributes import input_valid_attributes


@xml_root
def parse_input_xml(root):
    """ Parse an input.xml file into dictionary. """
    assert root.tag == "input"
    return parse_element_xml(root)


def get_root_from_tag(root: ElementTree.Element, tag: str = None) -> Tuple[ElementTree.Element, str]:
    """ Get the root from a tag.

    :param tag: tag of interest
    :param root: xml root containing the tag (or having the specified tag as tag)
    :returns: the tag and the found root, if tag was None returns the tag of the given root
    """
    if tag is None:
        return root, root.tag

    root = find_element(root, tag)
    if root is None:
        raise ValueError(f"Your specified input has no tag {tag}.")

    return root, root.tag


@xml_root
def parse_element_xml(root, tag: str = None) -> dict:
    """ Parse a xml element into dictionary. Can be input.xml root or a subelement of it.
    Put the attributes simply in dict and add recursively the subtrees and nested dicts.

    :param tag: the tag to parse
    :param root: the xml root containing the tag
    :returns: the parsed dictionary, data converted to actual data types
    """
    root, tag = get_root_from_tag(root, tag)

    if tag in special_tags_to_parse_map.keys():
        return special_tags_to_parse_map[tag](root)

    element_dict = convert_string_dict(copy.deepcopy(root.attrib))

    multiple_children = set(getattr(all_valid_attributes, f"{tag}_multiple_children", set()))
    subelements = list(root)
    multiple_tags = set([subelement.tag for subelement in subelements]) & multiple_children
    element_dict.update({tag: [] for tag in multiple_tags})

    # for loop over all subelements to retain the order
    for subelement in subelements:
        if subelement.tag in multiple_children:
            element_dict[subelement.tag].append(parse_element_xml(subelement))
        else:
            element_dict[subelement.tag] = parse_element_xml(subelement)

    return element_dict


def _parse_input_tag(root) -> dict:
    """ Parse special input tag. Necessary because exciting/xml allows arbitrary attributes at this level.
    Only parses the explicitly named attributes in the schema.

    :param root: the xml root containing the input tag
    :returns: the parsed dictionary, data converted to actual data types
    """
    valid_attribs = {key: value for key, value in root.attrib.items() if key in input_valid_attributes}
    element_dict = convert_string_dict(valid_attribs)

    subelements = list(root)
    for subelement in subelements:
        element_dict[subelement.tag] = parse_element_xml(subelement)

    return element_dict


@xml_root
def parse_structure(root) -> dict:
    """ Parse exciting input.xml structure element into python dictionary.

    :param root: Input for the parser.
    :returns: Dictionary containing the structure input element attributes and subelements. Looks like:
        {'atoms': List of atoms with atom positions in fractional coordinates,
         'lattice': List of 3 lattice vectors, 'species_path': species_path as string,
         'crystal_properties': dictionary with the crystal_properties,
         'species_properties': dictionary with the species_properties,
         all additional keys are structure attributes}
    """
    structure = find_element(root, 'structure')
    structure_properties = convert_string_dict(copy.deepcopy(structure.attrib))
    species_path = structure_properties.pop('speciespath')

    crystal = structure.find('crystal')
    crystal_properties = convert_string_dict(copy.deepcopy(crystal.attrib))
    lattice = []
    for base_vect in crystal:
        lattice.append([float(x) for x in base_vect.text.split()])

    atoms = []
    species_properties = {}
    for species in structure.findall('species'):
        species_attributes = convert_string_dict(copy.deepcopy(species.attrib))
        species_file = species_attributes.pop('speciesfile')
        species_symbol = species_file[:-4]

        species_subelements = list(species)
        atom_xml_trees = [x for x in species_subelements if x.tag == "atom"]
        for atom in atom_xml_trees:
            atom_attributes = convert_string_dict(copy.deepcopy(atom.attrib))
            atom_dict = {'species': species_symbol, 'position': atom_attributes.pop('coord')}
            atom_dict.update(atom_attributes)
            atoms.append(atom_dict)

        other_xml_trees = set(species_subelements) - set(atom_xml_trees)
        for tree in other_xml_trees:
            species_attributes[tree.tag] = parse_element_xml(tree)
        species_properties[species_symbol] = species_attributes

    return {
        'atoms': atoms,
        'lattice': lattice,
        'species_path': species_path,
        'crystal_properties': crystal_properties,
        'species_properties': species_properties,
        **structure_properties
    }


# special tag to parse function map or lambda if one-liner
# necessary for tags which doesn't contain simply xml attributes and subtrees
special_tags_to_parse_map = {"input": _parse_input_tag,
                             "title": lambda root: root.text,
                             "structure": parse_structure,
                             "qpointset": lambda root: [[float(x) for x in qpoint.text.split()] for qpoint in root],
                             "plan": lambda root: [doonly.attrib['task'] for doonly in root]}
